from __future__ import annotations

import json
import uuid
from typing import Optional, Self

import strawberry
from strawberry import ID, UNSET, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import to_global_id
from ai.backend.manager.services.object_storage.actions.get_download_presigned_url import (
    GetDownloadPresignedURLAction,
)
from ai.backend.manager.services.object_storage.actions.get_upload_presigned_url import (
    GetUploadPresignedURLAction,
)
from ai.backend.manager.services.storage_namespace.actions.get_multi import (
    GetNamespacesAction,
)

from ...data.object_storage.creator import ObjectStorageCreator
from ...data.object_storage.modifier import ObjectStorageModifier
from ...data.object_storage.types import ObjectStorageData
from ...services.object_storage.actions.create import CreateObjectStorageAction
from ...services.object_storage.actions.delete import DeleteObjectStorageAction
from ...services.object_storage.actions.get import GetObjectStorageAction
from ...services.object_storage.actions.list import ListObjectStorageAction
from ...services.object_storage.actions.update import UpdateObjectStorageAction
from ...types import OptionalState
from .storage_namespace import StorageNamespace, StorageNamespaceConnection, StorageNamespaceEdge
from .types import StrawberryGQLContext


@strawberry.type(description="Added in 25.14.0")
class ObjectStorage(Node):
    id: NodeID[str]
    name: str
    host: str
    access_key: str
    secret_key: str
    endpoint: str
    region: str

    @classmethod
    def from_dataclass(cls, data: ObjectStorageData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            host=data.host,
            access_key=data.access_key,
            secret_key=data.secret_key,
            endpoint=data.endpoint,
            region=data.region,
        )

    @strawberry.field
    async def namespaces(
        self,
        info: Info[StrawberryGQLContext],
        before: Optional[str],
        after: Optional[str],
        first: Optional[int],
        last: Optional[int],
        limit: Optional[int],
        offset: Optional[int],
    ) -> StorageNamespaceConnection:
        # TODO: Support pagination
        action_result = (
            await info.context.processors.storage_namespace.get_namespaces.wait_for_complete(
                GetNamespacesAction(uuid.UUID(self.id))
            )
        )

        nodes = [StorageNamespace.from_dataclass(bucket) for bucket in action_result.result]
        edges = [
            StorageNamespaceEdge(node=node, cursor=to_global_id(StorageNamespace, node.id))
            for node in nodes
        ]

        return StorageNamespaceConnection(
            edges=edges,
            page_info=strawberry.relay.PageInfo(
                has_next_page=False,
                has_previous_page=False,
                start_cursor=edges[0].cursor if edges else None,
                end_cursor=edges[-1].cursor if edges else None,
            ),
        )


ObjectStorageEdge = Edge[ObjectStorage]


@strawberry.type(description="Added in 25.14.0")
class ObjectStorageConnection(Connection[ObjectStorage]):
    @strawberry.field
    def count(self) -> int:
        return len(self.edges)


@strawberry.field(description="Added in 25.14.0")
async def object_storage(id: ID, info: Info[StrawberryGQLContext]) -> Optional[ObjectStorage]:
    processors = info.context.processors
    action_result = await processors.object_storage.get.wait_for_complete(
        GetObjectStorageAction(storage_id=uuid.UUID(id))
    )
    return ObjectStorage.from_dataclass(action_result.result)


@strawberry.field(description="Added in 25.14.0")
async def object_storages(
    info: Info[StrawberryGQLContext],
    before: Optional[str] = None,
    after: Optional[str] = None,
    first: Optional[int] = None,
    last: Optional[int] = None,
    limit: Optional[int] = None,
    offset: Optional[int] = None,
) -> ObjectStorageConnection:
    # TODO: Does we need to support filtering, ordering here?
    processors = info.context.processors

    action_result = await processors.object_storage.list_storages.wait_for_complete(
        ListObjectStorageAction()
    )

    nodes = [ObjectStorage.from_dataclass(data) for data in action_result.data]
    edges = [
        ObjectStorageEdge(node=node, cursor=to_global_id(ObjectStorage, node.id)) for node in nodes
    ]

    return ObjectStorageConnection(
        edges=edges,
        page_info=strawberry.relay.PageInfo(
            has_next_page=False,
            has_previous_page=False,
            start_cursor=edges[0].cursor if edges else None,
            end_cursor=edges[-1].cursor if edges else None,
        ),
    )


@strawberry.input(description="Added in 25.14.0")
class CreateObjectStorageInput:
    name: str
    host: str
    access_key: str
    secret_key: str
    endpoint: str
    region: str

    def to_creator(self) -> ObjectStorageCreator:
        return ObjectStorageCreator(
            name=self.name,
            host=self.host,
            access_key=self.access_key,
            secret_key=self.secret_key,
            endpoint=self.endpoint,
            region=self.region,
        )


@strawberry.input(description="Added in 25.14.0")
class UpdateObjectStorageInput:
    id: ID
    name: Optional[str] = UNSET
    host: Optional[str] = UNSET
    access_key: Optional[str] = UNSET
    secret_key: Optional[str] = UNSET
    endpoint: Optional[str] = UNSET
    region: Optional[str] = UNSET

    def to_modifier(self) -> ObjectStorageModifier:
        return ObjectStorageModifier(
            name=OptionalState[str].from_graphql(self.name),
            host=OptionalState[str].from_graphql(self.host),
            access_key=OptionalState[str].from_graphql(self.access_key),
            secret_key=OptionalState[str].from_graphql(self.secret_key),
            endpoint=OptionalState[str].from_graphql(self.endpoint),
            region=OptionalState[str].from_graphql(self.region),
        )


@strawberry.input(description="Added in 25.14.0")
class DeleteObjectStorageInput:
    id: ID


@strawberry.input(description="Added in 25.14.0")
class GetPresignedDownloadURLInput:
    artifact_revision_id: ID
    key: str
    expiration: Optional[int] = None


@strawberry.input(description="Added in 25.14.0")
class GetPresignedUploadURLInput:
    artifact_revision_id: ID
    key: str


@strawberry.type(description="Added in 25.14.0")
class CreateObjectStoragePayload:
    object_storage: ObjectStorage


@strawberry.type(description="Added in 25.14.0")
class UpdateObjectStoragePayload:
    object_storage: ObjectStorage


@strawberry.type(description="Added in 25.14.0")
class DeleteObjectStoragePayload:
    id: ID


@strawberry.type(description="Added in 25.14.0")
class GetPresignedDownloadURLPayload:
    presigned_url: str


@strawberry.type(description="Added in 25.14.0")
class GetPresignedUploadURLPayload:
    presigned_url: str
    fields: str  # JSON string containing the form fields


@strawberry.mutation(description="Added in 25.14.0")
async def create_object_storage(
    input: CreateObjectStorageInput, info: Info[StrawberryGQLContext]
) -> CreateObjectStoragePayload:
    processors = info.context.processors

    action_result = await processors.object_storage.create.wait_for_complete(
        CreateObjectStorageAction(
            creator=input.to_creator(),
        )
    )

    return CreateObjectStoragePayload(
        object_storage=ObjectStorage.from_dataclass(action_result.result)
    )


@strawberry.mutation(description="Added in 25.14.0")
async def update_object_storage(
    input: UpdateObjectStorageInput, info: Info[StrawberryGQLContext]
) -> UpdateObjectStoragePayload:
    processors = info.context.processors

    action_result = await processors.object_storage.update.wait_for_complete(
        UpdateObjectStorageAction(
            id=uuid.UUID(input.id),
            modifier=input.to_modifier(),
        )
    )

    return UpdateObjectStoragePayload(
        object_storage=ObjectStorage.from_dataclass(action_result.result)
    )


@strawberry.mutation(description="Added in 25.14.0")
async def delete_object_storage(
    input: DeleteObjectStorageInput, info: Info[StrawberryGQLContext]
) -> DeleteObjectStoragePayload:
    processors = info.context.processors

    action_result = await processors.object_storage.delete.wait_for_complete(
        DeleteObjectStorageAction(
            storage_id=uuid.UUID(input.id),
        )
    )

    return DeleteObjectStoragePayload(id=ID(str(action_result.deleted_storage_id)))


@strawberry.mutation(description="Added in 25.14.0")
async def get_presigned_download_url(
    input: GetPresignedDownloadURLInput, info: Info[StrawberryGQLContext]
) -> GetPresignedDownloadURLPayload:
    processors = info.context.processors

    action_result = await processors.object_storage.get_presigned_download_url.wait_for_complete(
        GetDownloadPresignedURLAction(
            artifact_revision_id=uuid.UUID(input.artifact_revision_id),
            key=input.key,
            expiration=input.expiration,
        )
    )

    return GetPresignedDownloadURLPayload(presigned_url=action_result.presigned_url)


@strawberry.mutation(description="Added in 25.14.0")
async def get_presigned_upload_url(
    input: GetPresignedUploadURLInput, info: Info[StrawberryGQLContext]
) -> GetPresignedUploadURLPayload:
    processors = info.context.processors

    action_result = await processors.object_storage.get_presigned_upload_url.wait_for_complete(
        GetUploadPresignedURLAction(
            artifact_revision_id=uuid.UUID(input.artifact_revision_id),
            key=input.key,
        )
    )

    return GetPresignedUploadURLPayload(
        presigned_url=action_result.presigned_url, fields=json.dumps(action_result.fields)
    )



from dataclasses import dataclass
from datetime import datetime
from typing import Optional, override
from uuid import UUID

from ai.backend.common.data.notification import NotifiableMessage, NotificationRuleType
from ai.backend.manager.actions.action import BaseActionResult

from .base import NotificationAction


@dataclass
class ProcessedRuleSuccess:
    """Information about a successfully processed notification rule."""

    rule_id: UUID
    rule_name: str
    channel_name: str


@dataclass
class ProcessNotificationAction(NotificationAction):
    """Action to process a notification event."""

    rule_type: NotificationRuleType
    timestamp: datetime
    notification_data: NotifiableMessage

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.rule_type)

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "process_notification"


@dataclass
class ProcessNotificationActionResult(BaseActionResult):
    """Result of processing a notification."""

    rule_type: NotificationRuleType
    rules_matched: int
    successes: list[ProcessedRuleSuccess]
    errors: list[BaseException]

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.rule_type)



"""Mount name validation rule."""

from typing import Any

from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.models import verify_vfolder_name
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    SessionCreationContext,
    SessionCreationSpec,
)

from .base import SessionValidatorRule


class MountNameValidationRule(SessionValidatorRule):
    """Validates vfolder mount names and aliases."""

    def name(self) -> str:
        return "mount_name_validation"

    def validate(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
    ) -> None:
        """Validate mount names if mount map is provided."""
        mount_map = spec.creation_spec.get("mount_map") or {}
        mount_id_map = spec.creation_spec.get("mount_id_map") or {}
        combined_mount_map = {**mount_map, **mount_id_map}

        if combined_mount_map:
            self._validate_mount_names(combined_mount_map)

    def _validate_mount_names(
        self,
        mount_map: dict[str, Any],
    ) -> None:
        """Validate mount alias names."""
        original_folders = list(mount_map.keys())
        alias_folders = list(mount_map.values())

        # Check for duplicate aliases
        if len(alias_folders) != len(set(alias_folders)):
            raise InvalidAPIParameters("Duplicate alias folder name exists.")

        for alias_name in alias_folders:
            if alias_name is None:
                continue

            # Remove work directory prefix if present
            if alias_name.startswith("/home/work/"):
                alias_name = alias_name.replace("/home/work/", "")

            # Validate alias name
            if alias_name == "":
                raise InvalidAPIParameters("Alias name cannot be empty.")

            if not verify_vfolder_name(alias_name):
                raise InvalidAPIParameters(f"{alias_name} is reserved for internal path.")

            # Check if alias conflicts with original folder names
            if alias_name in original_folders:
                raise InvalidAPIParameters(
                    f"Alias name cannot be set to an existing folder name: {alias_name}"
                )



import logging

from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.repositories.project_resource_policy.repository import (
    ProjectResourcePolicyRepository,
)
from ai.backend.manager.services.project_resource_policy.actions.create_project_resource_policy import (
    CreateProjectResourcePolicyAction,
    CreateProjectResourcePolicyActionResult,
)
from ai.backend.manager.services.project_resource_policy.actions.delete_project_resource_policy import (
    DeleteProjectResourcePolicyAction,
    DeleteProjectResourcePolicyActionResult,
)
from ai.backend.manager.services.project_resource_policy.actions.modify_project_resource_policy import (
    ModifyProjectResourcePolicyAction,
    ModifyProjectResourcePolicyActionResult,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ProjectResourcePolicyService:
    _project_resource_policy_repository: ProjectResourcePolicyRepository

    def __init__(
        self,
        project_resource_policy_repository: ProjectResourcePolicyRepository,
    ) -> None:
        self._project_resource_policy_repository = project_resource_policy_repository

    async def create_project_resource_policy(
        self, action: CreateProjectResourcePolicyAction
    ) -> CreateProjectResourcePolicyActionResult:
        fields = action.creator.fields_to_store()
        result = await self._project_resource_policy_repository.create(fields)
        return CreateProjectResourcePolicyActionResult(project_resource_policy=result)

    async def modify_project_resource_policy(
        self, action: ModifyProjectResourcePolicyAction
    ) -> ModifyProjectResourcePolicyActionResult:
        name = action.name
        to_update = action.modifier.fields_to_update()
        result = await self._project_resource_policy_repository.update(name, to_update)
        return ModifyProjectResourcePolicyActionResult(project_resource_policy=result)

    async def delete_project_resource_policy(
        self, action: DeleteProjectResourcePolicyAction
    ) -> DeleteProjectResourcePolicyActionResult:
        name = action.name
        result = await self._project_resource_policy_repository.delete(name)
        return DeleteProjectResourcePolicyActionResult(project_resource_policy=result)



"""Deployment service processors for GraphQL API."""

from typing import Protocol, override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor import ActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.deployment.actions.access_token.create_access_token import (
    CreateAccessTokenAction,
    CreateAccessTokenActionResult,
)
from ai.backend.manager.services.deployment.actions.access_token.list_access_tokens import (
    ListAccessTokensAction,
    ListAccessTokensActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.batch_load_auto_scaling_rules import (
    BatchLoadAutoScalingRulesAction,
    BatchLoadAutoScalingRulesActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.create_auto_scaling_rule import (
    CreateAutoScalingRuleAction,
    CreateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.delete_auto_scaling_rule import (
    DeleteAutoScalingRuleAction,
    DeleteAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.auto_scaling_rule.update_auto_scaling_rule import (
    UpdateAutoScalingRuleAction,
    UpdateAutoScalingRuleActionResult,
)
from ai.backend.manager.services.deployment.actions.batch_load_deployments import (
    BatchLoadDeploymentsAction,
    BatchLoadDeploymentsActionResult,
)
from ai.backend.manager.services.deployment.actions.batch_load_replicas_by_revision_ids import (
    BatchLoadReplicasByRevisionIdsAction,
    BatchLoadReplicasByRevisionIdsActionResult,
)
from ai.backend.manager.services.deployment.actions.create_deployment import (
    CreateDeploymentAction,
    CreateDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.create_legacy_deployment import (
    CreateLegacyDeploymentAction,
    CreateLegacyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.destroy_deployment import (
    DestroyDeploymentAction,
    DestroyDeploymentActionResult,
)
from ai.backend.manager.services.deployment.actions.list_deployments import (
    ListDeploymentsAction,
    ListDeploymentsActionResult,
)
from ai.backend.manager.services.deployment.actions.list_replicas import (
    ListReplicasAction,
    ListReplicasActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.add_model_revision import (
    AddModelRevisionAction,
    AddModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.batch_load_revisions import (
    BatchLoadRevisionsAction,
    BatchLoadRevisionsActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.create_model_revision import (
    CreateModelRevisionAction,
    CreateModelRevisionActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_deployment_id import (
    GetRevisionByDeploymentIdAction,
    GetRevisionByDeploymentIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_id import (
    GetRevisionByIdAction,
    GetRevisionByIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revision_by_replica_id import (
    GetRevisionByReplicaIdAction,
    GetRevisionByReplicaIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.get_revisions_by_deployment_id import (
    GetRevisionsByDeploymentIdAction,
    GetRevisionsByDeploymentIdActionResult,
)
from ai.backend.manager.services.deployment.actions.model_revision.list_revisions import (
    ListRevisionsAction,
    ListRevisionsActionResult,
)
from ai.backend.manager.services.deployment.actions.sync_replicas import (
    SyncReplicaAction,
    SyncReplicaActionResult,
)
from ai.backend.manager.services.deployment.actions.update_deployment import (
    UpdateDeploymentAction,
    UpdateDeploymentActionResult,
)


class DeploymentServiceProtocol(Protocol):
    async def create_deployment(
        self, action: CreateDeploymentAction
    ) -> CreateDeploymentActionResult: ...

    async def create_legacy_deployment(
        self, action: CreateLegacyDeploymentAction
    ) -> CreateLegacyDeploymentActionResult: ...

    async def update_deployment(
        self, action: UpdateDeploymentAction
    ) -> UpdateDeploymentActionResult: ...

    async def destroy_deployment(
        self, action: DestroyDeploymentAction
    ) -> DestroyDeploymentActionResult: ...

    async def batch_load_deployments(
        self, action: BatchLoadDeploymentsAction
    ) -> BatchLoadDeploymentsActionResult: ...

    async def create_auto_scaling_rule(
        self, action: CreateAutoScalingRuleAction
    ) -> CreateAutoScalingRuleActionResult: ...

    async def update_auto_scaling_rule(
        self, action: UpdateAutoScalingRuleAction
    ) -> UpdateAutoScalingRuleActionResult: ...

    async def delete_auto_scaling_rule(
        self, action: DeleteAutoScalingRuleAction
    ) -> DeleteAutoScalingRuleActionResult: ...

    async def create_access_token(
        self, action: CreateAccessTokenAction
    ) -> CreateAccessTokenActionResult: ...

    async def sync_replicas(self, action: SyncReplicaAction) -> SyncReplicaActionResult: ...

    async def add_model_revision(
        self, action: AddModelRevisionAction
    ) -> AddModelRevisionActionResult: ...

    async def batch_load_auto_scaling_rules(
        self, action: BatchLoadAutoScalingRulesAction
    ) -> BatchLoadAutoScalingRulesActionResult: ...

    async def get_revision_by_deployment_id(
        self, action: GetRevisionByDeploymentIdAction
    ) -> GetRevisionByDeploymentIdActionResult: ...

    async def get_revision_by_replica_id(
        self, action: GetRevisionByReplicaIdAction
    ) -> GetRevisionByReplicaIdActionResult: ...

    async def get_revision_by_id(
        self, action: GetRevisionByIdAction
    ) -> GetRevisionByIdActionResult: ...

    async def get_revisions_by_deployment_id(
        self, action: GetRevisionsByDeploymentIdAction
    ) -> GetRevisionsByDeploymentIdActionResult: ...

    async def batch_load_replicas_by_revision_ids(
        self, action: BatchLoadReplicasByRevisionIdsAction
    ) -> BatchLoadReplicasByRevisionIdsActionResult: ...

    async def batch_load_revisions(
        self, action: BatchLoadRevisionsAction
    ) -> BatchLoadRevisionsActionResult: ...

    async def list_replicas(self, action: ListReplicasAction) -> ListReplicasActionResult: ...
    async def list_revisions(self, action: ListRevisionsAction) -> ListRevisionsActionResult: ...

    async def create_model_revision(
        self, action: CreateModelRevisionAction
    ) -> CreateModelRevisionActionResult: ...

    async def list_access_tokens(
        self, action: ListAccessTokensAction
    ) -> ListAccessTokensActionResult: ...


class DeploymentProcessors(AbstractProcessorPackage):
    """Processors for deployment operations."""

    create_deployment: ActionProcessor[CreateDeploymentAction, CreateDeploymentActionResult]
    update_deployment: ActionProcessor[UpdateDeploymentAction, UpdateDeploymentActionResult]
    destroy_deployment: ActionProcessor[DestroyDeploymentAction, DestroyDeploymentActionResult]
    create_legacy_deployment: ActionProcessor[
        CreateLegacyDeploymentAction, CreateLegacyDeploymentActionResult
    ]
    create_auto_scaling_rule: ActionProcessor[
        CreateAutoScalingRuleAction, CreateAutoScalingRuleActionResult
    ]
    update_auto_scaling_rule: ActionProcessor[
        UpdateAutoScalingRuleAction, UpdateAutoScalingRuleActionResult
    ]
    delete_auto_scaling_rule: ActionProcessor[
        DeleteAutoScalingRuleAction, DeleteAutoScalingRuleActionResult
    ]
    create_access_token: ActionProcessor[CreateAccessTokenAction, CreateAccessTokenActionResult]
    list_access_tokens: ActionProcessor[ListAccessTokensAction, ListAccessTokensActionResult]
    sync_replicas: ActionProcessor[SyncReplicaAction, SyncReplicaActionResult]
    add_model_revision: ActionProcessor[AddModelRevisionAction, AddModelRevisionActionResult]
    batch_load_auto_scaling_rules: ActionProcessor[
        BatchLoadAutoScalingRulesAction, BatchLoadAutoScalingRulesActionResult
    ]
    get_revision_by_id: ActionProcessor[GetRevisionByIdAction, GetRevisionByIdActionResult]
    batch_load_revisions: ActionProcessor[BatchLoadRevisionsAction, BatchLoadRevisionsActionResult]
    get_revision_by_deployment_id: ActionProcessor[
        GetRevisionByDeploymentIdAction, GetRevisionByDeploymentIdActionResult
    ]
    get_revision_by_replica_id: ActionProcessor[
        GetRevisionByReplicaIdAction, GetRevisionByReplicaIdActionResult
    ]
    list_deployments: ActionProcessor[ListDeploymentsAction, ListDeploymentsActionResult]
    batch_load_deployments: ActionProcessor[
        BatchLoadDeploymentsAction, BatchLoadDeploymentsActionResult
    ]
    get_revisions_by_deployment_id: ActionProcessor[
        GetRevisionsByDeploymentIdAction, GetRevisionsByDeploymentIdActionResult
    ]
    batch_load_replicas_by_revision_ids: ActionProcessor[
        BatchLoadReplicasByRevisionIdsAction, BatchLoadReplicasByRevisionIdsActionResult
    ]
    list_replicas: ActionProcessor[ListReplicasAction, ListReplicasActionResult]
    list_revisions: ActionProcessor[ListRevisionsAction, ListRevisionsActionResult]
    create_model_revision: ActionProcessor[
        CreateModelRevisionAction, CreateModelRevisionActionResult
    ]

    def __init__(
        self, service: DeploymentServiceProtocol, action_monitors: list[ActionMonitor]
    ) -> None:
        self.create_auto_scaling_rule = ActionProcessor(
            service.create_auto_scaling_rule, action_monitors
        )
        self.update_auto_scaling_rule = ActionProcessor(
            service.update_auto_scaling_rule, action_monitors
        )
        self.delete_auto_scaling_rule = ActionProcessor(
            service.delete_auto_scaling_rule, action_monitors
        )
        self.batch_load_deployments = ActionProcessor(
            service.batch_load_deployments, action_monitors
        )
        self.create_deployment = ActionProcessor(service.create_deployment, action_monitors)
        self.destroy_deployment = ActionProcessor(service.destroy_deployment, action_monitors)
        self.update_deployment = ActionProcessor(service.update_deployment, action_monitors)
        self.create_legacy_deployment = ActionProcessor(
            service.create_legacy_deployment, action_monitors
        )
        self.create_access_token = ActionProcessor(service.create_access_token, action_monitors)
        self.list_access_tokens = ActionProcessor(service.list_access_tokens, action_monitors)
        self.sync_replicas = ActionProcessor(service.sync_replicas, action_monitors)
        self.add_model_revision = ActionProcessor(service.add_model_revision, action_monitors)
        self.batch_load_auto_scaling_rules = ActionProcessor(
            service.batch_load_auto_scaling_rules, action_monitors
        )
        self.get_revision_by_replica_id = ActionProcessor(
            service.get_revision_by_replica_id, action_monitors
        )
        self.get_revision_by_id = ActionProcessor(service.get_revision_by_id, action_monitors)
        self.get_revisions_by_deployment_id = ActionProcessor(
            service.get_revisions_by_deployment_id, action_monitors
        )
        self.batch_load_replicas_by_revision_ids = ActionProcessor(
            service.batch_load_replicas_by_revision_ids, action_monitors
        )
        self.list_replicas = ActionProcessor(service.list_replicas, action_monitors)
        self.list_revisions = ActionProcessor(service.list_revisions, action_monitors)
        self.create_model_revision = ActionProcessor(service.create_model_revision, action_monitors)
        self.batch_load_revisions = ActionProcessor(service.batch_load_revisions, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateDeploymentAction.spec(),
            DestroyDeploymentAction.spec(),
            CreateAutoScalingRuleAction.spec(),
            UpdateAutoScalingRuleAction.spec(),
            UpdateDeploymentAction.spec(),
            DeleteAutoScalingRuleAction.spec(),
            CreateAccessTokenAction.spec(),
            SyncReplicaAction.spec(),
            AddModelRevisionAction.spec(),
            BatchLoadAutoScalingRulesAction.spec(),
            GetRevisionByDeploymentIdAction.spec(),
            GetRevisionByReplicaIdAction.spec(),
            GetRevisionByIdAction.spec(),
            GetRevisionsByDeploymentIdAction.spec(),
            ListRevisionsAction.spec(),
            ListReplicasAction.spec(),
            CreateLegacyDeploymentAction.spec(),
            CreateModelRevisionAction.spec(),
            BatchLoadRevisionsAction.spec(),
            BatchLoadDeploymentsAction.spec(),
            ListAccessTokensAction.spec(),
            BatchLoadReplicasByRevisionIdsAction.spec(),
        ]



from dataclasses import dataclass, field
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.domain.types import DomainData, DomainModifier, UserInfo
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class ModifyDomainAction(DomainAction):
    domain_name: str
    user_info: UserInfo
    modifier: DomainModifier = field(default_factory=DomainModifier)

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "modify"

    def get_modified_fields(self):
        return self.modifier.fields_to_update()


@dataclass
class ModifyDomainActionResult(BaseActionResult):
    domain_data: Optional[DomainData]
    success: bool = field(compare=False)
    description: Optional[str] = field(compare=False)

    @override
    def entity_id(self) -> Optional[str]:
        return self.domain_data.name if self.domain_data is not None else None



from __future__ import annotations

import logging
import uuid
from http import HTTPStatus
from typing import TYPE_CHECKING, Iterable, Optional, Tuple

import aiohttp_cors
import sqlalchemy as sa
from aiohttp import web
from pydantic import Field
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.asyncio import AsyncSession

from ai.backend.common.api_handlers import BaseFieldModel
from ai.backend.common.container_registry import (
    ContainerRegistryType,
    PatchContainerRegistryRequestModel,
    PatchContainerRegistryResponseModel,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.container_registry.harbor import HarborRegistry_v2
from ai.backend.manager.errors.image import (
    ContainerRegistryWebhookAuthorizationFailed,
)
from ai.backend.manager.models.container_registry import (
    ContainerRegistryRow,
)
from ai.backend.manager.models.gql_models.container_registry_v2 import handle_allowed_groups_update

from ..errors.common import (
    GenericBadRequest,
    InternalServerError,
    ObjectNotFound,
)
from ..errors.image import HarborWebhookContainerRegistryRowNotFound

if TYPE_CHECKING:
    from .context import RootContext

from .auth import superadmin_required
from .manager import ALL_ALLOWED, READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import LegacyBaseRequestModel, pydantic_params_api_handler

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@server_status_required(READ_ALLOWED)
@superadmin_required
@pydantic_params_api_handler(PatchContainerRegistryRequestModel)
async def patch_container_registry(
    request: web.Request, params: PatchContainerRegistryRequestModel
) -> PatchContainerRegistryResponseModel:
    from ..models.container_registry import ContainerRegistryRow

    registry_id = uuid.UUID(request.match_info["registry_id"])
    log.info("PATCH_CONTAINER_REGISTRY (registry:{})", registry_id)
    root_ctx: RootContext = request.app["_root.context"]
    registry_row_updates = params.model_dump(exclude={"allowed_groups"}, exclude_none=True)

    try:
        async with root_ctx.db.begin_session() as db_session:
            if registry_row_updates:
                update_stmt = (
                    sa.update(ContainerRegistryRow)
                    .where(ContainerRegistryRow.id == registry_id)
                    .values(registry_row_updates)
                )
                await db_session.execute(update_stmt)

            query = sa.select(ContainerRegistryRow).where(ContainerRegistryRow.id == registry_id)
            container_registry = (await db_session.execute(query)).fetchone()[0]

        if params.allowed_groups:
            await handle_allowed_groups_update(root_ctx.db, registry_id, params.allowed_groups)
    except ObjectNotFound as e:
        raise e
    except IntegrityError as e:
        raise GenericBadRequest(f"Failed to update allowed groups! Details: {str(e)}")
    except Exception as e:
        raise InternalServerError(f"Failed to update container registry! Details: {str(e)}")

    return PatchContainerRegistryResponseModel.model_validate(container_registry)


async def _get_registry_row_matching_url(
    db_sess: AsyncSession, registry_url: str, project: str
) -> ContainerRegistryRow:
    query = sa.select(ContainerRegistryRow).where(
        (ContainerRegistryRow.type == ContainerRegistryType.HARBOR2)
        & (ContainerRegistryRow.url.like(f"%{registry_url}%"))
        & (ContainerRegistryRow.project == project)
    )
    result = await db_sess.execute(query)
    return result.scalars().one_or_none()


def _is_authorized_harbor_webhook_request(
    auth_header: Optional[str], registry_row: ContainerRegistryRow
) -> bool:
    if auth_header:
        extra = registry_row.extra or {}
        return extra.get("webhook_auth_header") == auth_header
    return True


async def _handle_harbor_webhook_event(
    root_ctx: RootContext,
    event_type: str,
    registry_row: ContainerRegistryRow,
    project: str,
    img_name: str,
    tag: str,
) -> None:
    match event_type:
        # Perform image rescan only for events that require it.
        case "PUSH_ARTIFACT":
            await _handle_push_artifact_event(root_ctx, registry_row, project, img_name, tag)
        case _:
            log.debug(
                'Ignore harbor webhook event: "{}". Recommended to modify the webhook config to not subscribe to this event type.',
                event_type,
            )


async def _handle_push_artifact_event(
    root_ctx: RootContext, registry_row: ContainerRegistryRow, project: str, img_name: str, tag: str
) -> None:
    scanner = HarborRegistry_v2(root_ctx.db, registry_row.registry_name, registry_row)
    await scanner.scan_single_ref(f"{project}/{img_name}:{tag}")


class HarborWebhookRequestModel(LegacyBaseRequestModel):
    type: str = Field(
        description="Type of the webhook event triggered by Harbor. See Harbor documentation for details."
    )

    class EventData(BaseFieldModel):
        class Resource(BaseFieldModel):
            resource_url: str = Field(description="URL of the artifact")
            tag: str = Field(description="Tag of the artifact")

        class Repository(BaseFieldModel):
            namespace: str = Field(description="Harbor project (namespace)")
            name: str = Field(description="Name of the repository")

        resources: list[Resource] = Field(
            description="List of related artifacts involved in the event"
        )
        repository: Repository = Field(description="Repository details")

    event_data: EventData = Field(description="Event details")


@server_status_required(ALL_ALLOWED)
@pydantic_params_api_handler(HarborWebhookRequestModel)
async def harbor_webhook_handler(
    request: web.Request, params: HarborWebhookRequestModel
) -> web.Response:
    auth_header = request.headers.get("Authorization", None)
    event_type = params.type
    resources = params.event_data.resources
    project = params.event_data.repository.namespace
    img_name = params.event_data.repository.name
    log.info("HARBOR_WEBHOOK_HANDLER (event_type:{})", event_type)

    root_ctx: RootContext = request.app["_root.context"]
    async with root_ctx.db.begin_session() as db_sess:
        for resource in resources:
            resource_url = resource.resource_url
            registry_url = resource_url.split("/")[0]

            registry_row = await _get_registry_row_matching_url(db_sess, registry_url, project)
            if not registry_row:
                raise HarborWebhookContainerRegistryRowNotFound(
                    extra_msg=f"Harbor webhook triggered, but the matching container registry row not found! (registry_url: {registry_url}, project: {project})",
                )

            if not _is_authorized_harbor_webhook_request(auth_header, registry_row):
                raise ContainerRegistryWebhookAuthorizationFailed(
                    extra_msg=f"Unauthorized webhook request (registry: {registry_row.registry_name}, project: {project})",
                )

            await _handle_harbor_webhook_event(
                root_ctx, event_type, registry_row, project, img_name, resource.tag
            )

    return web.Response(status=HTTPStatus.NO_CONTENT)


def create_app(
    default_cors_options: CORSOptions,
) -> Tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app["api_versions"] = (1, 2, 3, 4, 5)
    app["prefix"] = "container-registries"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("PATCH", "/{registry_id}", patch_container_registry))
    cors.add(app.router.add_route("POST", "/webhook/harbor", harbor_webhook_handler))
    return app, []



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.image.types import ImageData
from ai.backend.manager.services.image.actions.base import ImageAction


@dataclass
class ClearImageCustomResourceLimitAction(ImageAction):
    image_canonical: str
    architecture: str

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "clear_image_custom_resource_limit"


@dataclass
class ClearImageCustomResourceLimitActionResult(BaseActionResult):
    image_data: ImageData

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.image_data.id)



import uuid
from dataclasses import dataclass
from typing import Optional, override

from aiohttp import web

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.auth.actions.base import AuthAction


@dataclass
class SignupAction(AuthAction):
    request: web.Request
    domain_name: str
    email: str
    password: str
    username: Optional[str]
    full_name: Optional[str]
    description: Optional[str]

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "signup"

    @property
    def hook_params(self) -> dict[str, str]:
        params = {
            "domain": self.domain_name,
            "email": self.email,
            "password": self.password,
        }
        if self.username:
            params["username"] = self.username
        if self.full_name:
            params["full_name"] = self.full_name
        if self.description:
            params["description"] = self.description
        return params


@dataclass
class SignupActionResult(BaseActionResult):
    user_id: uuid.UUID
    access_key: str
    secret_key: str

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.user_id)



"""add vfolders domain name

Revision ID: d7df3baf3779
Revises: 5d92c9cc930c
Create Date: 2024-06-11 16:39:26.648608

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "d7df3baf3779"
down_revision = "5d92c9cc930c"
branch_labels = None
depends_on = None

DEFAULT_DOMAIN_NAME = "default"


def upgrade():
    conn = op.get_bind()
    op.add_column("vfolders", sa.Column("domain_name", sa.String(length=64), nullable=True))

    conn.execute(
        text(
            f"""\
        UPDATE vfolders
        SET domain_name = COALESCE(
            (SELECT domain_name FROM users WHERE vfolders.user = users.uuid),
            (SELECT domain_name FROM groups WHERE vfolders.group = groups.id),
            '{DEFAULT_DOMAIN_NAME}'
        )
        WHERE domain_name IS NULL;
    """
        )
    )

    op.alter_column("vfolders", column_name="domain_name", nullable=False)
    op.create_index(op.f("ix_vfolders_domain_name"), "vfolders", ["domain_name"], unique=False)


def downgrade():
    op.drop_index(op.f("ix_vfolders_domain_name"), table_name="vfolders")
    op.drop_column("vfolders", "domain_name")



from typing import Optional
from uuid import UUID

import sqlalchemy as sa
from sqlalchemy.ext.asyncio import AsyncConnection as SAConnection
from sqlalchemy.ext.asyncio import AsyncSession as SASession

from ai.backend.common.types import AccessKey

from .models import (
    UserRole,
    domains,
    groups,
    keypair_resource_policies,
    keypairs,
    users,
)
from .models import association_groups_users as agus


def check_if_requester_is_eligible_to_act_as_target_user(
    requester_role: UserRole,
    requester_domain: str,
    target_role: UserRole,
    target_domain: str,
) -> bool:
    if requester_role == UserRole.SUPERADMIN:
        pass
    elif requester_role == UserRole.ADMIN:
        if requester_domain != target_domain:
            raise RuntimeError(
                "Domain-admins can perform operations on behalf of "
                "other users in the same domain only.",
            )
        if target_role == UserRole.SUPERADMIN:
            raise RuntimeError(
                "Domain-admins cannot perform operations on behalf of super-admins.",
            )
        pass
    else:
        raise RuntimeError(
            "Only admins can perform operations on behalf of other users.",
        )
    return True


async def check_if_requester_is_eligible_to_act_as_target_access_key(
    conn: SAConnection,
    requester_role: UserRole,
    requester_domain: str,
    target_access_key: AccessKey,
) -> bool:
    query = (
        sa.select([users.c.domain_name, users.c.role])
        .select_from(sa.join(keypairs, users, keypairs.c.user == users.c.uuid))
        .where(keypairs.c.access_key == target_access_key)
    )
    result = await conn.execute(query)
    row = result.first()
    if row is None:
        raise ValueError("Unknown owner access key")
    owner_domain = row["domain_name"]
    owner_role = row["role"]
    return check_if_requester_is_eligible_to_act_as_target_user(
        requester_role,
        requester_domain,
        owner_role,
        owner_domain,
    )


async def check_if_requester_is_eligible_to_act_as_target_user_uuid(
    conn: SAConnection,
    requester_role: UserRole,
    requester_domain: str,
    target_user_uuid: UUID,
) -> bool:
    query = (
        sa.select([users.c.domain_name, users.c.role])
        .select_from(users)
        .where(users.c.uuid == target_user_uuid)
    )
    result = await conn.execute(query)
    row = result.first()
    if row is None:
        raise ValueError("Unknown owner access key")
    owner_domain = row["domain_name"]
    owner_role = row["role"]
    return check_if_requester_is_eligible_to_act_as_target_user(
        requester_role,
        requester_domain,
        owner_role,
        owner_domain,
    )


async def query_userinfo(
    conn: SAConnection,
    requester_uuid: UUID,
    requester_access_key: AccessKey,
    requester_role: UserRole,
    requester_domain: str,
    keypair_resource_policy: dict | None,
    requesting_domain: str,
    requesting_group: str | UUID,
    query_on_behalf_of: Optional[AccessKey] = None,
) -> tuple[UUID, UUID, dict]:
    if query_on_behalf_of is not None and query_on_behalf_of != requester_access_key:
        await check_if_requester_is_eligible_to_act_as_target_access_key(
            conn,
            requester_role,
            requester_domain,
            query_on_behalf_of,
        )
        owner_access_key = query_on_behalf_of
    else:
        owner_access_key = requester_access_key

    owner_uuid = None
    group_id = None
    resource_policy = None

    if keypair_resource_policy is None or requester_access_key != owner_access_key:
        # Admin or superadmin is creating sessions for another user.
        # The check for admin privileges is already done in get_access_key_scope().
        query = (
            sa.select([
                keypairs.c.user,
                keypairs.c.resource_policy,
                users.c.role,
                users.c.domain_name,
            ])
            .select_from(sa.join(keypairs, users, keypairs.c.user == users.c.uuid))
            .where(keypairs.c.access_key == owner_access_key)
        )
        result = await conn.execute(query)
        row = result.first()
        owner_domain = row["domain_name"]
        owner_uuid = row["user"]
        owner_role = row["role"]
        query = (
            sa.select([keypair_resource_policies])
            .select_from(keypair_resource_policies)
            .where(keypair_resource_policies.c.name == row["resource_policy"])
        )
        result = await conn.execute(query)
        resource_policy = result.first()
    else:
        # Normal case when the user is creating her/his own session.
        owner_domain = requester_domain
        owner_uuid = requester_uuid
        owner_role = UserRole.USER
        resource_policy = keypair_resource_policy

    query = (
        sa.select([domains.c.name])
        .select_from(domains)
        .where(
            (domains.c.name == owner_domain) & (domains.c.is_active),
        )
    )
    qresult = await conn.execute(query)
    domain_name = qresult.scalar()
    if domain_name is None:
        raise ValueError("Invalid domain")

    if isinstance(requesting_group, str):
        group_match_query = groups.c.name == requesting_group
    else:
        group_match_query = groups.c.id == requesting_group
    if owner_role == UserRole.SUPERADMIN:
        # superadmin can spawn container in any designated domain/group.
        query = (
            sa.select([groups.c.id])
            .select_from(groups)
            .where(
                (groups.c.domain_name == requesting_domain)
                & (group_match_query)
                & (groups.c.is_active),
            )
        )
        qresult = await conn.execute(query)
        group_id = qresult.scalar()
    elif owner_role == UserRole.ADMIN:
        # domain-admin can spawn container in any group in the same domain.
        if requesting_domain != owner_domain:
            raise ValueError("You can only set the domain to the owner's domain.")
        query = (
            sa.select([groups.c.id])
            .select_from(groups)
            .where(
                (groups.c.domain_name == owner_domain) & (group_match_query) & (groups.c.is_active),
            )
        )
        qresult = await conn.execute(query)
        group_id = qresult.scalar()
    else:
        # normal users can spawn containers in their group and domain.
        if requesting_domain != owner_domain:
            raise ValueError("You can only set the domain to your domain.")
        query = (
            sa.select([agus.c.group_id])
            .select_from(agus.join(groups, agus.c.group_id == groups.c.id))
            .where(
                (agus.c.user_id == owner_uuid)
                & (groups.c.domain_name == owner_domain)
                & (group_match_query)
                & (groups.c.is_active),
            )
        )
        qresult = await conn.execute(query)
        group_id = qresult.scalar()
    if group_id is None:
        raise ValueError("Invalid group")

    return owner_uuid, group_id, resource_policy


async def query_userinfo_from_session(
    db_sess: SASession,
    requester_uuid: UUID,
    requester_access_key: AccessKey,
    requester_role: UserRole,
    requester_domain: str,
    keypair_resource_policy: dict | None,
    requesting_domain: str,
    requesting_group: str | UUID,
    query_on_behalf_of: Optional[AccessKey] = None,
) -> tuple[UUID, UUID, dict]:
    """Version of query_userinfo that accepts SASession instead of SAConnection."""
    if query_on_behalf_of is not None and query_on_behalf_of != requester_access_key:
        # Need to check privileges - convert session operations
        query = (
            sa.select([users.c.domain_name, users.c.role])
            .select_from(sa.join(keypairs, users, keypairs.c.user == users.c.uuid))
            .where(keypairs.c.access_key == query_on_behalf_of)
        )
        result = await db_sess.execute(query)
        row = result.first()
        if row is None:
            raise ValueError("Unknown owner access key")
        owner_domain = row["domain_name"]
        owner_role = row["role"]
        check_if_requester_is_eligible_to_act_as_target_user(
            requester_role,
            requester_domain,
            owner_role,
            owner_domain,
        )
        owner_access_key = query_on_behalf_of
    else:
        owner_access_key = requester_access_key

    owner_uuid = None
    group_id = None
    resource_policy = None

    if keypair_resource_policy is None or requester_access_key != owner_access_key:
        # Admin or superadmin is creating sessions for another user.
        # The check for admin privileges is already done in get_access_key_scope().
        query = (
            sa.select([
                keypairs.c.user,
                keypairs.c.resource_policy,
                users.c.role,
                users.c.domain_name,
            ])
            .select_from(sa.join(keypairs, users, keypairs.c.user == users.c.uuid))
            .where(keypairs.c.access_key == owner_access_key)
        )
        result = await db_sess.execute(query)
        row = result.first()
        owner_domain = row["domain_name"]
        owner_uuid = row["user"]
        owner_role = row["role"]
        query = (
            sa.select([keypair_resource_policies])
            .select_from(keypair_resource_policies)
            .where(keypair_resource_policies.c.name == row["resource_policy"])
        )
        result = await db_sess.execute(query)
        resource_policy = result.first()
    else:
        # Normal case when the user is creating her/his own session.
        owner_domain = requester_domain
        owner_uuid = requester_uuid
        owner_role = UserRole.USER
        resource_policy = keypair_resource_policy

    query = (
        sa.select([domains.c.name])
        .select_from(domains)
        .where(
            (domains.c.name == owner_domain) & (domains.c.is_active),
        )
    )
    qresult = await db_sess.execute(query)
    domain_name = qresult.scalar()
    if domain_name is None:
        raise ValueError("Invalid domain")

    if isinstance(requesting_group, str):
        group_match_query = groups.c.name == requesting_group
    else:
        group_match_query = groups.c.id == requesting_group
    if owner_role == UserRole.SUPERADMIN:
        # superadmin can spawn container in any designated domain/group.
        query = (
            sa.select([groups.c.id])
            .select_from(groups)
            .where(
                (groups.c.domain_name == requesting_domain)
                & (group_match_query)
                & (groups.c.is_active),
            )
        )
        qresult = await db_sess.execute(query)
        group_id = qresult.scalar()
    elif owner_role == UserRole.ADMIN:
        # domain-admin can spawn container in any group in the same domain.
        if requesting_domain != owner_domain:
            raise ValueError("You can only set the domain to the owner's domain.")
        query = (
            sa.select([groups.c.id])
            .select_from(groups)
            .where(
                (groups.c.domain_name == owner_domain) & (group_match_query) & (groups.c.is_active),
            )
        )
        qresult = await db_sess.execute(query)
        group_id = qresult.scalar()
    else:
        # normal users can spawn containers in their group and domain.
        if requesting_domain != owner_domain:
            raise ValueError("You can only set the domain to your domain.")
        query = (
            sa.select([agus.c.group_id])
            .select_from(agus.join(groups, agus.c.group_id == groups.c.id))
            .where(
                (agus.c.user_id == owner_uuid)
                & (groups.c.domain_name == owner_domain)
                & (group_match_query)
                & (groups.c.is_active),
            )
        )
        qresult = await db_sess.execute(query)
        group_id = qresult.scalar()
    if group_id is None:
        raise ValueError("Invalid group")

    return owner_uuid, group_id, resource_policy



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

from ai.backend.common.types import (
    ResourceSlot,
)

from ..errors.common import GenericBadRequest

if TYPE_CHECKING:
    from ..models.session import SessionRow


def get_slot_index(slotname: str, agent_selection_resource_priority: list[str]) -> int:
    try:
        return agent_selection_resource_priority.index(slotname)
    except ValueError:
        return sys.maxsize


def sort_requested_slots_by_priority(
    requested_slots: ResourceSlot, agent_selection_resource_priority: list[str]
) -> list[str]:
    """
    Sort ``requested_slots``'s keys by the given ``agent_selection_resource_priority`` list.
    """

    for requested_slot_key in sorted(requested_slots.data.keys(), reverse=True):
        device_name = requested_slot_key.split(".")[0]
        if (
            requested_slot_key not in agent_selection_resource_priority
            and device_name in agent_selection_resource_priority
        ):
            agent_selection_resource_priority.insert(
                agent_selection_resource_priority.index(device_name) + 1, requested_slot_key
            )

    return sorted(
        requested_slots.data.keys(),
        key=lambda item: get_slot_index(item, agent_selection_resource_priority),
    )


def get_requested_architecture(sess_ctx: SessionRow) -> str:
    requested_architectures = set(k.architecture for k in sess_ctx.kernels)
    if len(requested_architectures) > 1:
        raise GenericBadRequest(
            "Cannot assign multiple kernels with different architectures' single node session",
        )
    return requested_architectures.pop()



"""
WebSocket-based streaming kernel interaction APIs.
"""

from __future__ import annotations

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Callable, Coroutine, Optional, TypeVar, Union

import aiohttp
import aiotools
from aiohttp import WSCloseCode, web

from ai.backend.logging import BraceStyleAdapter

from ..config_legacy import DEFAULT_CHUNK_SIZE

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

TArg = TypeVar("TArg")
TRet = TypeVar("TRet")
type AsyncCallback[TArg, TRet] = Callable[[TArg], Coroutine[Any, Any, TRet]]


class ServiceProxy(metaclass=ABCMeta):
    """
    The abstract base class to implement service proxy handlers.
    """

    __slots__ = (
        "ws",
        "host",
        "port",
        "downstream_cb",
        "upstream_cb",
        "ping_cb",
    )

    def __init__(
        self,
        down_ws: web.WebSocketResponse,
        dest_host: str,
        dest_port: int,
        *,
        downstream_callback: Optional[AsyncCallback[Any, None]] = None,
        upstream_callback: Optional[AsyncCallback[Any, None]] = None,
        ping_callback: Optional[AsyncCallback[Any, None]] = None,
    ) -> None:
        self.ws = down_ws
        self.host = dest_host
        self.port = dest_port
        self.downstream_cb = downstream_callback
        self.upstream_cb = upstream_callback
        self.ping_cb = ping_callback

    @abstractmethod
    async def proxy(self) -> web.WebSocketResponse:
        pass


class TCPProxy(ServiceProxy):
    __slots__ = (
        *ServiceProxy.__slots__,
        "down_task",
    )

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.down_task: Optional[asyncio.Task] = None

    async def proxy(self) -> web.WebSocketResponse:
        try:
            try:
                log.debug("Trying to open proxied TCP connection to {}:{}", self.host, self.port)
                reader, writer = await asyncio.open_connection(self.host, self.port)
            except ConnectionRefusedError:
                await self.ws.close(code=WSCloseCode.TRY_AGAIN_LATER)
                return self.ws
            except Exception:
                log.exception("TCPProxy.proxy(): unexpected initial connection error")
                await self.ws.close(code=WSCloseCode.INTERNAL_ERROR)
                return self.ws

            async def downstream() -> None:
                try:
                    while True:
                        try:
                            chunk = await reader.read(DEFAULT_CHUNK_SIZE)
                            if not chunk:
                                break
                            await self.ws.send_bytes(chunk)
                        except (RuntimeError, ConnectionResetError, asyncio.CancelledError):
                            # connection interrupted by client-side
                            break
                        else:
                            if self.downstream_cb is not None:
                                await self.downstream_cb(chunk)
                except asyncio.CancelledError:
                    pass
                except Exception:
                    log.exception("TCPProxy.proxy(): unexpected downstream error")
                finally:
                    await self.ws.close(code=WSCloseCode.GOING_AWAY)

            log.debug("TCPProxy connected {0}:{1}", self.host, self.port)
            self.down_task = asyncio.create_task(downstream())
            async for msg in self.ws:
                if msg.type == web.WSMsgType.BINARY:
                    try:
                        writer.write(msg.data)
                        await writer.drain()
                    except RuntimeError:
                        log.debug("Error on writing: Is it closed?")
                    if self.upstream_cb is not None:
                        await self.upstream_cb(msg.data)
                elif msg.type == web.WSMsgType.PING:
                    await self.ws.pong(msg.data)
                    if self.ping_cb is not None:
                        await self.ping_cb(msg.data)
                elif msg.type == web.WSMsgType.ERROR:
                    log.debug("TCPProxy.proxy(): websocket upstream error", exc_info=msg.data)
                    writer.close()
                    await writer.wait_closed()

        except asyncio.CancelledError:
            pass
        except Exception:
            log.exception("TCPProxy.proxy(): unexpected upstream error")
        finally:
            if self.down_task is not None and not self.down_task.done():
                self.down_task.cancel()
                await self.down_task
            log.debug("websocket connection closed")
        return self.ws


class WebSocketProxy:
    __slots__ = (
        "up_conn",
        "down_conn",
        "upstream_buffer",
        "upstream_buffer_task",
        "downstream_cb",
        "upstream_cb",
        "ping_cb",
    )

    up_conn: aiohttp.ClientWebSocketResponse
    down_conn: web.WebSocketResponse
    # FIXME: use __future__.annotations in Python 3.7+
    upstream_buffer: asyncio.Queue  # contains: Tuple[Union[bytes, str], web.WSMsgType]
    upstream_buffer_task: Optional[asyncio.Future[None]]
    downstream_cb: AsyncCallback[str | bytes, None] | None
    upstream_cb: AsyncCallback[str | bytes, None] | None
    ping_cb: AsyncCallback[str | bytes, None] | None

    def __init__(
        self,
        up_conn: aiohttp.ClientWebSocketResponse,
        down_conn: web.WebSocketResponse,
        *,
        downstream_callback: Optional[AsyncCallback[str | bytes, None]] = None,
        upstream_callback: Optional[AsyncCallback[str | bytes, None]] = None,
        ping_callback: Optional[AsyncCallback[str | bytes, None]] = None,
    ):
        self.up_conn = up_conn
        self.down_conn = down_conn
        self.upstream_buffer = asyncio.Queue()
        self.upstream_buffer_task = None
        self.downstream_cb = downstream_callback
        self.upstream_cb = upstream_callback
        self.ping_cb = ping_callback

    async def proxy(self) -> None:
        asyncio.create_task(self.downstream())
        await self.upstream()

    async def upstream(self) -> None:
        try:
            async for msg in self.down_conn:
                if msg.type in (web.WSMsgType.TEXT, web.WSMsgType.binary):
                    await self.write(msg.data, msg.type)
                    if self.upstream_cb is not None:
                        await self.upstream_cb(msg.data)
                elif msg.type == web.WSMsgType.PING:
                    if self.ping_cb is not None:
                        await self.ping_cb(msg.data)
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    log.error("ws connection closed with exception {}", self.up_conn.exception())
                    break
                elif msg.type == aiohttp.WSMsgType.CLOSE:
                    break
            # here, client gracefully disconnected
        except asyncio.CancelledError:
            # here, client forcibly disconnected
            raise
        finally:
            await self.close_downstream()

    async def downstream(self) -> None:
        try:
            async with aiotools.TaskScope() as ts:
                self.upstream_buffer_task = ts.create_task(self.consume_upstream_buffer())
                async for msg in self.up_conn:
                    if msg.type == aiohttp.WSMsgType.TEXT:
                        await self.down_conn.send_str(msg.data)
                        if self.downstream_cb is not None:
                            ts.create_task(self.downstream_cb(msg.data))
                    if msg.type == aiohttp.WSMsgType.BINARY:
                        await self.down_conn.send_bytes(msg.data)
                        if self.downstream_cb is not None:
                            ts.create_task(self.downstream_cb(msg.data))
                    elif msg.type == aiohttp.WSMsgType.CLOSED:
                        break
                    elif msg.type == aiohttp.WSMsgType.ERROR:
                        break
            # here, server gracefully disconnected
        except asyncio.CancelledError:
            raise
        except Exception:
            log.exception("unexpected error")
        finally:
            await self.close_upstream()

    async def consume_upstream_buffer(self) -> None:
        while True:
            msg, tp = await self.upstream_buffer.get()
            try:
                if self.up_conn and not self.up_conn.closed:
                    if tp == aiohttp.WSMsgType.TEXT:
                        await self.up_conn.send_str(msg)
                    elif tp == aiohttp.WSMsgType.binary:
                        await self.up_conn.send_bytes(msg)
                else:
                    await self.close_downstream()
            finally:
                self.upstream_buffer.task_done()

    async def write(self, msg: Union[bytes, str], tp: web.WSMsgType) -> None:
        await self.upstream_buffer.put((msg, tp))

    async def close_downstream(self) -> None:
        if not self.down_conn.closed:
            await self.down_conn.close()

    async def close_upstream(self) -> None:
        if self.upstream_buffer_task:
            self.upstream_buffer_task.cancel()
            await self.upstream_buffer_task
        if self.up_conn:
            await self.up_conn.close()



from __future__ import annotations

import logging
from contextlib import asynccontextmanager as actxmgr
from typing import Any, AsyncIterator, Mapping, Optional

from ai.backend.common.docker import ImageRef
from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience
from ai.backend.common.types import (
    AgentId,
    ClusterInfo,
    ImageConfig,
    KernelCreationConfig,
    KernelId,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.agent_cache import AgentRPCCache, PeerInvoker

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

agent_client_resilience = Resilience(
    policies=[
        MetricPolicy(MetricArgs(domain=DomainType.CLIENT, layer=LayerType.AGENT_CLIENT)),
        RetryPolicy(
            RetryArgs(
                max_retries=3,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.EXPONENTIAL,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class AgentClient:
    """
    Client for communicating with a single agent via RPC.
    Wraps all RPC calls to the agent to provide a cleaner interface.
    """

    _agent_cache: AgentRPCCache
    _agent_id: AgentId
    _invoke_timeout: Optional[float]
    _order_key: Optional[str]

    def __init__(
        self,
        agent_cache: AgentRPCCache,
        agent_id: AgentId,
        *,
        invoke_timeout: Optional[float] = None,
        order_key: Optional[str] = None,
    ) -> None:
        self._agent_cache = agent_cache
        self._agent_id = agent_id
        self._invoke_timeout = invoke_timeout
        self._order_key = order_key

    @property
    def agent_id(self) -> AgentId:
        return self._agent_id

    @actxmgr
    async def _with_connection(self) -> AsyncIterator[PeerInvoker]:
        """
        Context manager to get a PeerInvoker for the agent.
        This will automatically handle connection management.
        """
        async with self._agent_cache.rpc_context(
            self._agent_id,
            invoke_timeout=self._invoke_timeout,
            order_key=self._order_key,
        ) as rpc:
            yield rpc

    # Hardware information methods
    @agent_client_resilience.apply()
    async def health(self) -> Mapping[str, Any]:
        """Get lightweight health information from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.health()

    @agent_client_resilience.apply()
    async def gather_hwinfo(self) -> Mapping[str, Any]:
        """Gather hardware information from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.gather_hwinfo(agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def scan_gpu_alloc_map(self) -> Mapping[str, Any]:
        """Scan GPU allocation map from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.scan_gpu_alloc_map(agent_id=self.agent_id)

    # Image management methods
    @agent_client_resilience.apply()
    async def check_and_pull(self, image_configs: Mapping[str, ImageConfig]) -> Mapping[str, str]:
        """Check and pull images on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.check_and_pull(image_configs, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def purge_images(
        self,
        images: list[str],
        force: bool,
        noprune: bool,
    ) -> Mapping[str, Any]:
        """Purge images from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.purge_images(images, force, noprune, agent_id=self.agent_id)

    # Network management methods
    @agent_client_resilience.apply()
    async def create_local_network(self, network_name: str) -> None:
        """Create a local network on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.create_local_network(network_name, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def destroy_local_network(self, network_ref_name: str) -> None:
        """Destroy a local network on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.destroy_local_network(network_ref_name, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def assign_port(self) -> int:
        """Assign a host port on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.assign_port(agent_id=self.agent_id)

    # Kernel management methods
    @agent_client_resilience.apply()
    async def create_kernels(
        self,
        session_id: str,
        kernel_ids: list[str],
        kernel_configs: list[KernelCreationConfig],
        cluster_info: ClusterInfo,
        kernel_image_refs: Mapping[KernelId, ImageRef],
    ) -> Any:
        """Create kernels on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.create_kernels(
                session_id,
                kernel_ids,
                kernel_configs,
                cluster_info,
                kernel_image_refs,
                agent_id=self.agent_id,
            )

    @agent_client_resilience.apply()
    async def destroy_kernel(
        self,
        kernel_id: str,
        session_id: str,
        reason: str,
        suppress_events: bool = True,
    ) -> None:
        """Destroy a kernel on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.destroy_kernel(
                kernel_id,
                session_id,
                reason,
                suppress_events=suppress_events,
                agent_id=self.agent_id,
            )

    @agent_client_resilience.apply()
    async def restart_kernel(
        self,
        session_id: str,
        kernel_id: str,
        image_ref: Any,
        update_config: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        """Restart a kernel on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.restart_kernel(
                session_id,
                kernel_id,
                image_ref,
                update_config,
                agent_id=self.agent_id,
            )

    @agent_client_resilience.apply()
    async def sync_kernel_registry(self, kernel_tuples: list[tuple[str, str]]) -> None:
        """Sync kernel registry on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.sync_kernel_registry(kernel_tuples, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def drop_kernel_registry(self, kernel_id_list: list[KernelId]) -> None:
        """Drop kernel registry entries on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.drop_kernel_registry(kernel_id_list, agent_id=self.agent_id)

    # Health monitoring methods
    @agent_client_resilience.apply()
    async def check_pulling(self, image_name: str) -> bool:
        """Check if an image is being pulled."""
        async with self._with_connection() as rpc:
            return await rpc.call.check_pulling(image_name, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def check_creating(self, kernel_id: str) -> bool:
        """Check if a kernel is being created."""
        async with self._with_connection() as rpc:
            return await rpc.call.check_creating(str(kernel_id), agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def check_running(self, kernel_id: str) -> bool:
        """Check if a kernel is running."""
        async with self._with_connection() as rpc:
            return await rpc.call.check_running(str(kernel_id), agent_id=self.agent_id)

    # Container management methods
    @agent_client_resilience.apply()
    async def purge_containers(self, serialized_data: list[tuple[str, str]]) -> None:
        """Purge containers on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.purge_containers(serialized_data, agent_id=self.agent_id)

    # Code execution methods
    @agent_client_resilience.apply()
    async def execute(
        self,
        session_id: str,
        kernel_id: str,
        major_api_version: int,
        run_id: str,
        mode: str,
        code: str,
        opts: Mapping[str, Any],
        flush_timeout: float | None,
    ) -> Mapping[str, Any]:
        """Execute code on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.execute(
                session_id,
                kernel_id,
                major_api_version,
                run_id,
                mode,
                code,
                opts,
                flush_timeout,
                agent_id=self.agent_id,
            )

    @agent_client_resilience.apply()
    async def interrupt_kernel(self, kernel_id: str) -> Mapping[str, Any]:
        """Interrupt a kernel on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.interrupt_kernel(kernel_id, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def trigger_batch_execution(
        self,
        session_id: str,
        kernel_id: str,
        startup_command: str,
        batch_timeout: float,
    ) -> None:
        """Trigger batch execution on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.trigger_batch_execution(
                session_id,
                kernel_id,
                startup_command,
                batch_timeout,
                agent_id=self.agent_id,
            )

    @agent_client_resilience.apply()
    async def get_completions(
        self,
        kernel_id: str,
        text: str,
        opts: Mapping[str, Any],
    ) -> dict[str, Any]:
        """Get code completions from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.get_completions(kernel_id, text, opts, agent_id=self.agent_id)

    # Service management methods
    @agent_client_resilience.apply()
    async def start_service(
        self,
        kernel_id: str,
        service: str,
        opts: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        """Start a service on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.start_service(kernel_id, service, opts, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def shutdown_service(self, kernel_id: str, service: str) -> None:
        """Shutdown a service on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.shutdown_service(kernel_id, service, agent_id=self.agent_id)

    # File management methods
    @agent_client_resilience.apply()
    async def upload_file(self, kernel_id: str, filename: str, payload: bytes) -> Mapping[str, Any]:
        """Upload a file to the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.upload_file(kernel_id, filename, payload, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def download_file(self, kernel_id: str, filepath: str) -> bytes:
        """Download a file from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.download_file(kernel_id, filepath, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def download_single(self, kernel_id: str, filepath: str) -> bytes:
        """Download a single file from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.download_single(kernel_id, filepath, agent_id=self.agent_id)

    @agent_client_resilience.apply()
    async def list_files(self, kernel_id: str, path: str) -> Mapping[str, Any]:
        """List files on the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.list_files(kernel_id, path, agent_id=self.agent_id)

    # Log management methods
    @agent_client_resilience.apply()
    async def get_logs(self, kernel_id: str) -> Mapping[str, str]:
        """Get logs from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.get_logs(kernel_id, agent_id=self.agent_id)

    # Image commit methods
    @agent_client_resilience.apply()
    async def commit(
        self,
        kernel_id: str,
        email: str,
        canonical: Optional[str] = None,
        extra_labels: Optional[dict[str, str]] = None,
        filename: Optional[str] = None,
    ) -> Mapping[str, Any]:
        """Commit a kernel image on the agent."""
        async with self._with_connection() as rpc:
            kwargs: dict[str, Any] = {}
            if canonical is not None:
                kwargs["canonical"] = canonical
            if extra_labels is not None:
                kwargs["extra_labels"] = extra_labels
            if filename is not None:
                kwargs["filename"] = filename
            kwargs["agent_id"] = self.agent_id

            return await rpc.call.commit(kernel_id, email, **kwargs)

    @agent_client_resilience.apply()
    async def push_image(self, image_ref: ImageRef, registry: Any) -> Mapping[str, Any]:
        """Push an image from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.push_image(image_ref, registry, agent_id=self.agent_id)

    # Scaling group management
    @agent_client_resilience.apply()
    async def update_scaling_group(self, scaling_group: str) -> None:
        """Update scaling group on the agent."""
        async with self._with_connection() as rpc:
            await rpc.call.update_scaling_group(scaling_group, self.agent_id)

    # Local configuration management
    @agent_client_resilience.apply()
    async def get_local_config(self) -> Mapping[str, str]:
        """Get local configuration from the agent."""
        async with self._with_connection() as rpc:
            return await rpc.call.get_local_config(self.agent_id)



"""remove size limit for model_service_token.token

Revision ID: 4b0128c49210
Revises: 02535458c0b3
Create Date: 2023-09-01 10:55:44.074487

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import IDColumn

# revision identifiers, used by Alembic.
revision = "4b0128c49210"
down_revision = "02535458c0b3"
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint("pk_endpoint_tokens", "endpoint_tokens")
    op.add_column("endpoint_tokens", IDColumn())
    op.alter_column("endpoint_tokens", "token", type_=sa.String())


def downgrade():
    op.drop_column("pk_endpoint_tokens", "endpoint_tokens")
    op.drop_column("endpoint_tokens", "id")
    op.alter_column("endpoint_tokens", "token", type_=sa.VARCHAR(1024))
    op.create_primary_key("pk_endpoint_tokens", "endpoint_tokens", ["token"])



"""add uid and gid columns

Revision ID: f6ca2f2d04c1
Revises: fb89f5d7817b
Create Date: 2024-12-20 12:16:07.077845

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "f6ca2f2d04c1"
down_revision = "fb89f5d7817b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "kernels", sa.Column("uid", sa.Integer(), server_default=sa.text("NULL"), nullable=True)
    )
    op.add_column(
        "kernels",
        sa.Column("main_gid", sa.Integer(), server_default=sa.text("NULL"), nullable=True),
    )
    op.add_column(
        "kernels",
        sa.Column("gids", sa.ARRAY(sa.Integer()), server_default=sa.text("NULL"), nullable=True),
    )
    op.add_column(
        "users",
        sa.Column("container_uid", sa.Integer(), server_default=sa.text("NULL"), nullable=True),
    )
    op.add_column(
        "users",
        sa.Column(
            "container_main_gid", sa.Integer(), server_default=sa.text("NULL"), nullable=True
        ),
    )
    op.add_column(
        "users",
        sa.Column(
            "container_gids",
            sa.ARRAY(sa.Integer()),
            server_default=sa.text("NULL"),
            nullable=True,
        ),
    )


def downgrade() -> None:
    op.drop_column("users", "container_main_gid")
    op.drop_column("users", "container_gids")
    op.drop_column("users", "container_uid")
    op.drop_column("kernels", "gids")
    op.drop_column("kernels", "main_gid")
    op.drop_column("kernels", "uid")



import uuid
from collections.abc import Mapping
from typing import Optional

from ai.backend.common.exception import BackendAIError
from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience.policies.metrics import MetricArgs, MetricPolicy
from ai.backend.common.resilience.policies.retry import BackoffStrategy, RetryArgs, RetryPolicy
from ai.backend.common.resilience.resilience import Resilience

from ...data.permission.id import ObjectId
from ...data.permission.role import (
    BatchEntityPermissionCheckInput,
    RoleCreateInput,
    RoleData,
    RoleDeleteInput,
    RoleUpdateInput,
    ScopePermissionCheckInput,
    SingleEntityPermissionCheckInput,
    UserRoleAssignmentInput,
)
from ...models.utils import ExtendedAsyncSAEngine
from .db_source import PermissionDBSource

permission_controller_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(
                domain=DomainType.REPOSITORY, layer=LayerType.PERMISSION_CONTROLLER_REPOSITORY
            )
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
                non_retryable_exceptions=(BackendAIError,),
            )
        ),
    ]
)


class PermissionControllerRepository:
    _db_source: PermissionDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = PermissionDBSource(db)

    @permission_controller_repository_resilience.apply()
    async def create_role(self, data: RoleCreateInput) -> RoleData:
        """
        Create a new role in the database.

        Returns the ID of the created role.
        """
        role_row = await self._db_source.create_role(data)
        return role_row.to_data()

    @permission_controller_repository_resilience.apply()
    async def update_role(self, data: RoleUpdateInput) -> RoleData:
        result = await self._db_source.update_role(data)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def delete_role(self, data: RoleDeleteInput) -> RoleData:
        result = await self._db_source.delete_role(data)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def assign_role(self, data: UserRoleAssignmentInput):
        result = await self._db_source.assign_role(data)
        return result.to_data()

    @permission_controller_repository_resilience.apply()
    async def get_role(self, role_id: uuid.UUID) -> Optional[RoleData]:
        result = await self._db_source.get_role(role_id)
        return result.to_data() if result else None

    @permission_controller_repository_resilience.apply()
    async def check_permission_of_entity(self, data: SingleEntityPermissionCheckInput) -> bool:
        target_object_id = data.target_object_id
        roles = await self._db_source.get_user_roles(data.user_id)
        associated_scopes = await self._db_source.get_entity_mapped_scopes(target_object_id)
        associated_scopes_set = set([row.parsed_scope_id() for row in associated_scopes])
        for role in roles:
            for object_perm in role.object_permission_rows:
                if object_perm.operation != data.operation:
                    continue
                if object_perm.object_id() == target_object_id:
                    return True

            for permission_group in role.permission_group_rows:
                if permission_group.parsed_scope_id() not in associated_scopes_set:
                    continue
                for permission in permission_group.permission_rows:
                    if permission.operation == data.operation:
                        return True
        return False

    @permission_controller_repository_resilience.apply()
    async def check_permission_in_scope(self, data: ScopePermissionCheckInput) -> bool:
        return await self._db_source.check_scope_permission_exist(
            data.user_id, data.target_scope_id, data.operation
        )

    @permission_controller_repository_resilience.apply()
    async def check_permission_of_entities(
        self,
        data: BatchEntityPermissionCheckInput,
    ) -> Mapping[ObjectId, bool]:
        """
        Check if the user has the requested operation permission on the given entity IDs.
        Returns a mapping of entity ID to a boolean indicating permission.
        """
        return await self._db_source.check_batch_object_permission_exist(
            data.user_id, data.target_object_ids, data.operation
        )



from __future__ import annotations

import enum
import functools
import re
from collections.abc import (
    Awaitable,
    Callable,
)
from typing import (
    Any,
    Generic,
    NamedTuple,
    Protocol,
    TypeAlias,
    TypeVar,
)

import graphene
import graphql
from graphene.relay.connection import (
    ConnectionOptions,
    IterableConnectionField,
    PageInfo,
)
from graphene.relay.node import Node, NodeField, is_node
from graphene.types import Field, NonNull, ObjectType, String
from graphene.types.objecttype import ObjectTypeMeta
from graphene.types.utils import get_type
from graphql_relay.utils import base64, unbase64

from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.errors.common import ServerMisconfiguredError
from ai.backend.manager.errors.resource import DataTransformationFailed


def get_edge_class(
    connection_class: type[Connection],
    _node: type[AsyncNode],
    base_name: str,
    strict_types: bool = False,
    description: str | None = None,
) -> ObjectTypeMeta:
    edge_class = getattr(connection_class, "Edge", None)

    class EdgeBase:
        node = Field(
            NonNull(_node) if strict_types else _node,
            description="The item at the end of the edge",
        )
        cursor = String(required=True, description="A cursor for use in pagination")

    edge_description = f"A Relay edge containing a `{base_name}` and its cursor."
    if description is not None:
        edge_description = f"{description} {edge_description}"

    class EdgeMeta:
        description = edge_description

    edge_name = f"{base_name}Edge"

    edge_bases = [edge_class, EdgeBase] if edge_class else [EdgeBase]
    if not isinstance(edge_class, ObjectType):
        edge_bases = [*edge_bases, ObjectType]

    return type(edge_name, tuple(edge_bases), {"Meta": EdgeMeta})


class PageInfoType(Protocol):
    @property
    def start_cursor(self) -> str | None: ...

    @property
    def end_cursor(self) -> str | None: ...

    @property
    def has_previous_page(self) -> bool: ...

    @property
    def has_next_page(self) -> bool: ...


class EdgeType(Protocol):
    @property
    def node(self) -> Any: ...

    @property
    def cursor(self) -> str: ...


class ConnectionType(Protocol):
    @property
    def edges(self) -> list[EdgeType]: ...

    @property
    def page_info(self) -> PageInfoType: ...


class EdgeConstructor(Protocol):
    def __call__(
        self,
        *,
        node: Any,
        cursor: str,
    ) -> EdgeType: ...


class ConnectionConstructor(Protocol):
    @property
    def Edge(self) -> EdgeConstructor: ...

    def __call__(
        self,
        *,
        edges: list[EdgeType],
        page_info: PageInfoType,
        count: int,
    ) -> Connection: ...


class AsyncNodeField(NodeField):
    def wrap_resolve(self, parent_resolver):
        return functools.partial(self.node_type.node_resolver, get_type(self.field_type))


def _resolve_global_id(global_id: str) -> tuple[str, str]:
    unbased_global_id = unbase64(global_id)
    type_, _, id_ = unbased_global_id.partition(":")
    return type_, id_


class AsyncNode(Node):
    """
    This GraphQL Relay Node extension is for running asynchronous resolvers and fine-grained handling of global id.
    Refer to: https://github.com/graphql-python/graphene/blob/master/graphene/relay/node.py
    """

    class Meta:
        name = "Node"

    @classmethod
    def Field(cls, *args, **kwargs):
        return AsyncNodeField(cls, *args, **kwargs)

    @classmethod
    async def node_resolver(cls, only_type, root, info, id):
        return await cls.get_node_from_global_id(info, id, only_type=only_type)

    @staticmethod
    def to_global_id(type_, id_) -> str:
        if id_ is None:
            raise Exception("Encoding None value as Global ID is not allowed.")
        return base64(f"{type_}:{id_}")

    @classmethod
    def resolve_global_id(cls, info, global_id: str) -> tuple[str, str]:
        return _resolve_global_id(global_id)

    @classmethod
    async def get_node_from_global_id(cls, info, global_id: str, only_type=None) -> Any:
        _type, _ = cls.resolve_global_id(info, global_id)

        graphene_type = info.schema.get_type(_type)
        if graphene_type is None:
            raise Exception(f'Relay Node "{_type}" not found in schema')

        graphene_type = graphene_type.graphene_type

        if only_type:
            if graphene_type != only_type:
                raise InvalidAPIParameters(f"Must receive a {only_type._meta.name} id.")

        if cls not in graphene_type._meta.interfaces:
            raise Exception(f'ObjectType "{_type}" does not implement the "{cls}" interface.')

        get_node = getattr(graphene_type, "get_node", None)
        if get_node:
            return await get_node(info, global_id)
        raise Exception(f'ObjectType "{_type}" does not implement `get_node` method.')


class Connection(graphene.ObjectType):
    """
    This GraphQL Relay Connection has been implemented to have additional fields, such as `count`.
    Refer to: https://github.com/graphql-python/graphene/blob/master/graphene/relay/connection.py
    """

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass_with_meta__(
        cls, node=None, name=None, strict_types=False, _meta=None, **options
    ):
        if not _meta:
            _meta = ConnectionOptions(cls)
        if not node:
            raise ServerMisconfiguredError(f"You have to provide a node in {cls.__name__}.Meta")
        if not (
            isinstance(node, graphene.NonNull)
            or issubclass(
                node,
                (
                    graphene.Scalar,
                    graphene.Enum,
                    graphene.ObjectType,
                    graphene.Interface,
                    graphene.Union,
                    graphene.NonNull,
                ),
            )
        ):
            raise ServerMisconfiguredError(
                f'Received incompatible node "{node}" for Connection {cls.__name__}.'
            )

        base_name = re.sub("Connection$", "", name or cls.__name__) or node._meta.name
        if not name:
            name = f"{base_name}Connection"

        options["name"] = name

        _meta.node = node

        if not _meta.fields:
            _meta.fields = {}

        if "page_info" not in _meta.fields:
            _meta.fields["page_info"] = graphene.Field(
                PageInfo,
                name="pageInfo",
                required=True,
                description="Pagination data for this connection.",
            )

        if "edges" not in _meta.fields:
            connection_description = options.get("description")
            edge_class = get_edge_class(cls, node, base_name, strict_types, connection_description)  # type: ignore
            cls.Edge = edge_class
            _meta.fields["edges"] = graphene.Field(
                graphene.NonNull(
                    graphene.List(graphene.NonNull(edge_class) if strict_types else edge_class)
                ),
                description="Contains the nodes in this connection.",
            )

        if "count" not in _meta.fields:
            _meta.fields["count"] = graphene.Field(
                graphene.Int(),
                name="count",
                description="Total count of the GQL nodes of the query.",
            )

        return super().__init_subclass_with_meta__(_meta=_meta, **options)


class ConnectionPaginationOrder(enum.Enum):
    FORWARD = "forward"
    BACKWARD = "backward"


T_Node = TypeVar("T_Node", bound=ObjectType)


class ConnectionResolverResult(Generic[T_Node], NamedTuple):
    node_list: list[T_Node] | Connection
    cursor: str | None
    pagination_order: ConnectionPaginationOrder | None
    requested_page_size: int | None
    total_count: int


Resolver = (
    Callable[..., Awaitable[ConnectionResolverResult]] | Callable[..., ConnectionResolverResult]
)


class AsyncListConnectionField(IterableConnectionField):
    """
    This GraphQL Relay Connection field extension is for getting paginated list data from asynchronous resolvers.
    The resolver function of graphene.relay.Connection is implemented
    to accept only one complete array (Iterable values) without considering pagination, which is a huge performance issue.
    Refer to: https://github.com/graphql-python/graphene/blob/master/graphene/relay/connection.py
    """

    @property
    def type(self):
        type_ = super(IterableConnectionField, self).type
        connection_type = type_
        if isinstance(type_, graphene.NonNull):
            connection_type = type_.of_type

        if is_node(connection_type):
            raise Exception("ConnectionFields now need a explicit ConnectionType for Nodes.")

        if not issubclass(connection_type, Connection):
            raise ServerMisconfiguredError(
                f"{self.__class__.__name__} type has to be a subclass of"
                f' ai.backend.manager.models.gql_relay.Connection. Received "{connection_type}".'
            )
        return type_

    @classmethod
    def resolve_connection(
        cls,
        connection_type: ConnectionConstructor,
        args: dict[str, Any] | None,
        resolver_result: ConnectionResolverResult,
    ) -> Connection:
        resolved = resolver_result.node_list
        page_size = resolver_result.requested_page_size
        pagination_order = resolver_result.pagination_order
        count = resolver_result.total_count

        if isinstance(resolved, Connection):
            return resolved

        if not isinstance(resolved, list):
            raise DataTransformationFailed(
                "Resolved value from the connection field has to be a list or instance of"
                f' {connection_type}. Received "{resolved}"'
            )

        orig_resolved_len = len(resolved)
        if page_size is not None:
            resolved = resolved[:page_size]
        if pagination_order == ConnectionPaginationOrder.BACKWARD:
            resolved = resolved[::-1]
        edge_type = connection_type.Edge
        edges = [
            edge_type(
                node=value,
                cursor=AsyncNode.to_global_id(str(connection_type._meta.node), value.id),  # type: ignore[attr-defined]
            )
            for value in resolved
        ]
        return connection_type(
            edges=edges,
            page_info=PageInfo(
                start_cursor=edges[0].cursor if edges else None,
                end_cursor=edges[-1].cursor if edges else None,
                has_previous_page=(
                    pagination_order == ConnectionPaginationOrder.BACKWARD
                    and page_size is not None
                    and page_size < orig_resolved_len
                ),
                has_next_page=(
                    pagination_order == ConnectionPaginationOrder.FORWARD
                    and page_size is not None
                    and page_size < orig_resolved_len
                ),
            ),
            count=count,
        )

    @classmethod
    async def connection_resolver(
        cls,
        resolver: Resolver,
        connection_type: ConnectionConstructor,
        root,
        info,
        **args,
    ) -> Connection:
        _result = resolver(root, info, **args)
        match _result:
            case ConnectionResolverResult():
                result = _result
            case _:
                result = await _result

        if isinstance(connection_type, graphene.NonNull):
            connection_type = connection_type.of_type

        return cls.resolve_connection(
            connection_type,
            args,
            resolver_result=result,
        )


ConnectionField = AsyncListConnectionField

NodeTypeName: TypeAlias = str
NodeID: TypeAlias = str
ResolvedGlobalID: TypeAlias = tuple[NodeTypeName, NodeID]


def _from_str(value: str) -> ResolvedGlobalID:
    type_, id_ = _resolve_global_id(value)
    if not id_:
        return type_, value
    return type_, id_


class GlobalIDField(graphene.Scalar):
    class Meta:
        description = (
            "Added in 24.09.0. Global ID of GQL relay spec. "
            'Base64 encoded version of "<node type name>:<node id>". '
            "UUID or string type values are also allowed."
        )

    @staticmethod
    def serialize(val: Any) -> Any:
        return val

    @staticmethod
    def parse_literal(node: Any, _variables=None) -> ResolvedGlobalID | None:
        if isinstance(node, graphql.language.ast.StringValueNode):
            return _from_str(node.value)
        return None

    @staticmethod
    def parse_value(value: str) -> ResolvedGlobalID:
        return _from_str(value)



from .types import (
    NotificationChannelCreator,
    NotificationChannelData,
    NotificationChannelListResult,
    NotificationChannelModifier,
    NotificationChannelType,
    NotificationRuleCreator,
    NotificationRuleData,
    NotificationRuleListResult,
    NotificationRuleModifier,
    NotificationRuleType,
    WebhookConfig,
)

__all__ = (
    "NotificationChannelCreator",
    "NotificationChannelData",
    "NotificationChannelListResult",
    "NotificationChannelModifier",
    "NotificationChannelType",
    "NotificationRuleCreator",
    "NotificationRuleData",
    "NotificationRuleListResult",
    "NotificationRuleModifier",
    "NotificationRuleType",
    "WebhookConfig",
)



"""Session preparer that applies multiple preparation rules."""

import logging
import uuid
from collections.abc import Iterable
from datetime import datetime
from typing import Any, Optional

from dateutil.tz import tzutc

from ai.backend.common.types import (
    KernelEnqueueingConfig,
    KernelId,
    ResourceSlot,
    SessionId,
    VFolderMount,
)
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.data.kernel.types import KernelStatus
from ai.backend.manager.data.session.types import SessionStatus
from ai.backend.manager.defs import DEFAULT_ROLE
from ai.backend.manager.models.network import NetworkType
from ai.backend.manager.repositories.scheduler.types.session_creation import (
    AllowedScalingGroup,
    KernelEnqueueData,
    SessionCreationContext,
    SessionCreationSpec,
    SessionEnqueueData,
)

from ..types import CalculatedResources
from .base import SessionPreparerRule

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class SessionPreparer:
    """
    Preparer that applies multiple preparation rules for session creation.
    """

    _rules: Iterable[SessionPreparerRule]

    # Default values for missing data
    DEFAULT_IMAGE_NAME = "ai.backend.undefined"
    DEFAULT_ARCHITECTURE = "x86_64"
    DEFAULT_REGISTRY = "localhost"

    def __init__(self, rules: Iterable[SessionPreparerRule]) -> None:
        self._rules = rules

    async def prepare(
        self,
        spec: SessionCreationSpec,
        validated_scaling_group: AllowedScalingGroup,
        context: SessionCreationContext,
        calculated_resources: CalculatedResources,
    ) -> SessionEnqueueData:
        """
        Convert creation spec to enqueue data.

        Args:
            spec: Session creation specification
            validated_scaling_group: Validated scaling group
            context: Pre-fetched context with all required data
            calculated_resources: Pre-calculated resource information

        Returns:
            Complete SessionEnqueueData with kernels and dependencies
        """
        # Initialize preparation data
        preparation_data: dict[str, Any] = {
            "dotfile_data": context.dotfile_data,
            "vfolder_mounts": context.vfolder_mounts,
        }

        # Apply all preparation rules
        for rule in self._rules:
            log.debug(f"Applying preparation rule: {rule.name()}")
            rule.prepare(spec, context, preparation_data)

        # Generate session ID if not provided
        session_id = SessionId(uuid.uuid4())

        # Use network info from context
        scaling_group_network = context.scaling_group_network

        # Prepare environment variables
        environ = dict(spec.creation_spec.get("environ") or {})

        # Add SUDO_SESSION_ENABLED to environ if sudo session is enabled
        if spec.sudo_session_enabled:
            environ["SUDO_SESSION_ENABLED"] = "1"

        # Determine network configuration
        network_type, network_id = self._determine_network_config(spec, context)

        # Extract bootstrap script and timeout
        bootstrap_script = spec.creation_spec.get("bootstrap_script")
        timeout = spec.creation_spec.get("session_timeout")

        # Build kernel data using kernel_configs from preparation_data
        kernel_configs = preparation_data.get("kernel_configs", spec.kernel_specs)
        internal_data = preparation_data.get("internal_data", {})

        kernel_data_list = await self._prepare_kernels(
            spec,
            session_id,
            validated_scaling_group,
            kernel_configs,
            internal_data,
            context.vfolder_mounts,
            context,
            calculated_resources,
        )

        # Collect images from kernels
        session_images = self._collect_session_images(kernel_data_list)

        # Use pre-calculated session total
        total_requested = calculated_resources.session_requested_slots

        # Build complete session data with all information
        session_data = SessionEnqueueData(
            id=session_id,
            creation_id=spec.session_creation_id,
            name=spec.session_name,
            access_key=spec.access_key,
            user_uuid=spec.user_scope.user_uuid,
            group_id=spec.user_scope.group_id,
            domain_name=spec.user_scope.domain_name,
            scaling_group_name=validated_scaling_group.name,
            session_type=spec.session_type,
            cluster_mode=spec.cluster_mode,
            cluster_size=spec.cluster_size,
            priority=spec.priority,
            status=SessionStatus.PENDING.name,
            status_history={
                SessionStatus.PENDING.name: datetime.now(tzutc()).isoformat(),
            },
            requested_slots=total_requested,
            occupying_slots=ResourceSlot(),
            vfolder_mounts=context.vfolder_mounts,
            environ=environ,
            tag=spec.session_tag,
            starts_at=spec.starts_at,
            batch_timeout=(int(spec.batch_timeout.total_seconds()) if spec.batch_timeout else None),
            callback_url=spec.callback_url,
            images=session_images,
            network_type=network_type,
            network_id=network_id,
            bootstrap_script=bootstrap_script,
            designated_agent_list=self._get_designated_agent_ids(
                spec=spec, kernel_config=kernel_configs
            ),
            use_host_network=scaling_group_network.use_host_network,
            timeout=timeout,
            kernels=kernel_data_list,
            dependencies=spec.dependency_sessions or [],
        )

        return session_data

    def _determine_network_config(
        self,
        spec: SessionCreationSpec,
        context: SessionCreationContext,
    ) -> tuple[NetworkType | None, str | None]:
        """Determine network type and ID from spec."""
        if spec.network:
            return NetworkType.PERSISTENT, str(spec.network.id)
        elif context.scaling_group_network.use_host_network:
            return NetworkType.HOST, None
        else:
            # Default to VOLATILE for multi-container or single-container sessions
            return NetworkType.VOLATILE, None

    async def _prepare_kernels(
        self,
        spec: SessionCreationSpec,
        session_id: SessionId,
        validated_scaling_group: AllowedScalingGroup,
        kernel_configs: list[KernelEnqueueingConfig],
        internal_data: dict[str, Any],
        vfolder_mounts: list[VFolderMount],
        context: SessionCreationContext,
        calculated_resources: CalculatedResources,
    ) -> list[KernelEnqueueData]:
        """Prepare kernel enqueue data."""
        kernel_data_list = []

        for idx, kernel_config in enumerate(kernel_configs):
            kernel_id = KernelId(uuid.uuid4())

            # Get image info
            image_ref = kernel_config.get("image_ref")
            if image_ref and hasattr(image_ref, "canonical"):
                image_info = context.image_infos.get(image_ref.canonical)
            else:
                # Fallback for string image references
                image_info = None

            # Get preopen_ports from kernel config or fall back to creation_spec
            # This follows the same pattern as in the original registry.py
            preopen_ports = kernel_config.get("creation_config", {}).get("preopen_ports")
            if preopen_ports is None:
                # Fall back to creation_spec preopen_ports (applies to all kernels)
                preopen_ports = spec.creation_spec.get("preopen_ports")
            if not preopen_ports:
                preopen_ports = []
            if not isinstance(preopen_ports, list):
                preopen_ports = []

            # Use pre-calculated resources for this kernel
            if idx < len(calculated_resources.kernel_resources):
                kernel_resource = calculated_resources.kernel_resources[idx]
                requested_slots = kernel_resource.requested_slots
                resource_opts = kernel_resource.resource_opts
                log.debug(
                    "Kernel {} (idx={}, role={}) using calculated resources: {}",
                    kernel_id,
                    idx,
                    kernel_config.get("cluster_role", "unknown"),
                    requested_slots,
                )
            else:
                # Fallback if no calculated resources
                log.warning(
                    "Kernel {} (idx={}, role={}) has no calculated resources! "
                    "calculated_resources has {} entries but need idx={}",
                    kernel_id,
                    idx,
                    kernel_config.get("cluster_role", "unknown"),
                    len(calculated_resources.kernel_resources),
                    idx,
                )
                requested_slots = ResourceSlot()
                resource_opts = {}

            # Build kernel data
            kernel_data = KernelEnqueueData(
                id=kernel_id,
                session_id=session_id,
                session_creation_id=spec.session_creation_id,
                session_name=spec.session_name,
                session_type=spec.session_type,
                cluster_mode=spec.cluster_mode,
                cluster_size=spec.cluster_size,
                cluster_role=kernel_config.get("cluster_role", DEFAULT_ROLE),
                cluster_idx=kernel_config.get("cluster_idx", idx),
                local_rank=kernel_config.get("local_rank", idx),
                cluster_hostname=kernel_config.get("cluster_hostname")
                or f"{kernel_config.get('cluster_role', DEFAULT_ROLE)}{kernel_config.get('cluster_idx', idx + 1)}",
                scaling_group=validated_scaling_group.name,
                domain_name=spec.user_scope.domain_name,
                group_id=spec.user_scope.group_id,
                user_uuid=spec.user_scope.user_uuid,
                access_key=spec.access_key,
                image=image_info.canonical if image_info else self.DEFAULT_IMAGE_NAME,
                architecture=image_info.architecture if image_info else self.DEFAULT_ARCHITECTURE,
                registry=image_info.registry if image_info else self.DEFAULT_REGISTRY,
                tag=spec.session_tag,
                starts_at=spec.starts_at,
                status=KernelStatus.PENDING.name,
                status_history={
                    KernelStatus.PENDING.name: datetime.now(tzutc()).isoformat(),
                },
                occupied_slots=ResourceSlot(),
                requested_slots=requested_slots,
                occupied_shares={},
                resource_opts=resource_opts,
                environ=[f"{k}={v}" for k, v in (kernel_config.get("environ", {}) or {}).items()]  # type: ignore[attr-defined]
                if kernel_config.get("environ")
                else [],
                bootstrap_script=kernel_config.get("bootstrap_script"),
                startup_command=kernel_config.get("startup_command"),
                internal_data=internal_data,
                callback_url=spec.callback_url,
                mounts=[mount.name for mount in vfolder_mounts],  # Legacy field
                vfolder_mounts=vfolder_mounts,
                preopen_ports=preopen_ports if isinstance(preopen_ports, list) else [],
                use_host_network=context.scaling_group_network.use_host_network,
                uid=kernel_config.get("uid") or context.container_user_info.uid,
                main_gid=kernel_config.get("main_gid") or context.container_user_info.main_gid,
                gids=kernel_config.get("supplementary_gids")
                or context.container_user_info.supplementary_gids,
            )

            kernel_data_list.append(kernel_data)

        return kernel_data_list

    def _get_designated_agent_ids(
        self,
        spec: SessionCreationSpec,
        kernel_config: list[KernelEnqueueingConfig],
    ) -> Optional[list[str]]:
        """Get pre-assigned agent for a kernel."""
        # Check if agent is specified in kernel config
        designated_agents = set()
        for kernel_cfg in kernel_config:
            if agent_id := kernel_cfg.get("agent"):
                designated_agents.add(str(agent_id))
        if designated_agents:
            return list(designated_agents)
        return spec.designated_agent_list

    def _collect_session_images(
        self,
        kernel_data_list: list[KernelEnqueueData],
    ) -> list[str]:
        """Collect unique images from kernels, with main kernel first."""
        session_images: list[str] = []

        for kernel in kernel_data_list:
            if kernel.image not in session_images:
                if kernel.cluster_role == DEFAULT_ROLE:
                    # Main kernel image goes first
                    session_images.insert(0, kernel.image)
                else:
                    session_images.append(kernel.image)

        return session_images



from dataclasses import dataclass
from typing import Self

from ai.backend.manager.repositories.scheduler.repository import SchedulerRepository
from ai.backend.manager.repositories.types import RepositoryArgs


@dataclass
class SchedulerRepositories:
    repository: SchedulerRepository

    @classmethod
    def create(cls, args: RepositoryArgs) -> Self:
        repository = SchedulerRepository(args.db, args.valkey_stat_client, args.config_provider)

        return cls(
            repository=repository,
        )



from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import Optional

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.data.artifact_registry.types import HuggingFaceRegistryStatefulData


@dataclass
class HuggingFaceRegistryData:
    id: uuid.UUID
    name: str
    url: str
    token: Optional[str]

    @classmethod
    def from_stateful_data(
        cls, stateful_data: HuggingFaceRegistryStatefulData
    ) -> HuggingFaceRegistryData:
        """Convert HuggingFaceRegistryStatefulData to HuggingFaceRegistryData."""
        return cls(
            id=stateful_data.id,
            name=stateful_data.name,
            url=stateful_data.url,
            token=stateful_data.token,
        )

    def to_stateful_data(self, artifact_registry_id: uuid.UUID) -> HuggingFaceRegistryStatefulData:
        """Convert HuggingFaceRegistryData to HuggingFaceRegistryStatefulData for caching.

        :param artifact_registry_id: The UUID from artifact_registries table.
        """
        return HuggingFaceRegistryStatefulData(
            id=artifact_registry_id,
            registry_id=self.id,
            name=self.name,
            type=ArtifactRegistryType.HUGGINGFACE,
            url=self.url,
            token=self.token,
        )



"""Rename `object_storage_meta` to `object_storage_namespace`

Revision ID: b6f3b3c69d16
Revises: 29bbf3830309
Create Date: 2025-08-28 05:09:38.548556

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "b6f3b3c69d16"
down_revision = "29bbf3830309"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_storage_id_bucket", "object_storage_meta")

    op.create_table(
        "object_storage_namespace",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),  # noqa: F821
        sa.Column("storage_id", GUID(), nullable=False),
        sa.Column("bucket", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_object_storage_namespace")),
        sa.UniqueConstraint("storage_id", "bucket", name="uq_storage_id_bucket"),
    )

    # Migrate data from object_storage_meta to object_storage_namespace
    connection = op.get_bind()
    connection.execute(
        sa.text("""
            INSERT INTO object_storage_namespace (id, storage_id, bucket)
            SELECT id, storage_id, bucket FROM object_storage_meta
        """)
    )

    op.drop_table("object_storage_meta")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint("uq_storage_id_bucket", "object_storage_namespace")

    op.create_table(
        "object_storage_meta",
        sa.Column(
            "id",
            GUID(),
            server_default=sa.text("uuid_generate_v4()"),
            autoincrement=False,
            nullable=False,
        ),
        sa.Column("storage_id", GUID(), autoincrement=False, nullable=False),
        sa.Column("bucket", sa.String(), autoincrement=False, nullable=False),
        sa.PrimaryKeyConstraint("id", name="pk_object_storage_meta"),
        sa.UniqueConstraint("storage_id", "bucket", name="uq_storage_id_bucket"),
    )

    # Migrate data from object_storage_namespace to object_storage_meta
    connection = op.get_bind()
    connection.execute(
        sa.text("""
            INSERT INTO object_storage_meta (id, storage_id, bucket)
            SELECT id, storage_id, bucket FROM object_storage_namespace
        """)
    )

    op.drop_table("object_storage_namespace")
    # ### end Alembic commands ###



"""GraphQL types, filters, and inputs for notification system."""

from __future__ import annotations

import uuid
from datetime import datetime
from enum import StrEnum
from typing import Optional, Self, override

import strawberry
from strawberry import ID, UNSET
from strawberry.relay import Node, NodeID

from ai.backend.common.data.notification import (
    NotificationChannelType,
    NotificationRuleType,
    WebhookConfig,
)
from ai.backend.manager.api.gql.base import OrderDirection, StringFilter
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy
from ai.backend.manager.data.notification import (
    NotificationChannelCreator,
    NotificationChannelData,
    NotificationChannelModifier,
    NotificationRuleCreator,
    NotificationRuleData,
    NotificationRuleModifier,
)
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.notification.options import (
    NotificationChannelConditions,
    NotificationChannelOrders,
    NotificationRuleConditions,
    NotificationRuleOrders,
)
from ai.backend.manager.types import OptionalState

# GraphQL enum types


@strawberry.enum(name="NotificationChannelType", description="Notification channel types")
class NotificationChannelTypeGQL(StrEnum):
    WEBHOOK = "webhook"

    @classmethod
    def from_internal(cls, internal_type: NotificationChannelType) -> NotificationChannelTypeGQL:
        """Convert internal NotificationChannelType to GraphQL enum."""
        match internal_type:
            case NotificationChannelType.WEBHOOK:
                return cls.WEBHOOK

    def to_internal(self) -> NotificationChannelType:
        """Convert GraphQL enum to internal NotificationChannelType."""
        match self:
            case NotificationChannelTypeGQL.WEBHOOK:
                return NotificationChannelType.WEBHOOK


@strawberry.enum(name="NotificationRuleType", description="Notification rule types")
class NotificationRuleTypeGQL(StrEnum):
    SESSION_STARTED = "session.started"
    SESSION_TERMINATED = "session.terminated"
    ARTIFACT_DOWNLOAD_COMPLETED = "artifact.download.completed"

    @classmethod
    def from_internal(cls, internal_type: NotificationRuleType) -> NotificationRuleTypeGQL:
        """Convert internal NotificationRuleType to GraphQL enum."""
        match internal_type:
            case NotificationRuleType.SESSION_STARTED:
                return cls.SESSION_STARTED
            case NotificationRuleType.SESSION_TERMINATED:
                return cls.SESSION_TERMINATED
            case NotificationRuleType.ARTIFACT_DOWNLOAD_COMPLETED:
                return cls.ARTIFACT_DOWNLOAD_COMPLETED

    def to_internal(self) -> NotificationRuleType:
        """Convert GraphQL enum to internal NotificationRuleType."""
        match self:
            case NotificationRuleTypeGQL.SESSION_STARTED:
                return NotificationRuleType.SESSION_STARTED
            case NotificationRuleTypeGQL.SESSION_TERMINATED:
                return NotificationRuleType.SESSION_TERMINATED
            case NotificationRuleTypeGQL.ARTIFACT_DOWNLOAD_COMPLETED:
                return NotificationRuleType.ARTIFACT_DOWNLOAD_COMPLETED


# GraphQL object types


@strawberry.type
class WebhookConfigGQL:
    """GraphQL type for webhook configuration."""

    url: str

    @classmethod
    def from_dataclass(cls, config: WebhookConfig) -> Self:
        return cls(
            url=config.url,
        )


@strawberry.type(description="Notification channel")
class NotificationChannel(Node):
    id: NodeID[str]
    name: str
    description: Optional[str]
    channel_type: NotificationChannelTypeGQL
    config: WebhookConfigGQL
    enabled: bool
    created_at: datetime

    @classmethod
    def from_dataclass(cls, data: NotificationChannelData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            description=data.description,
            channel_type=NotificationChannelTypeGQL.from_internal(data.channel_type),
            config=WebhookConfigGQL.from_dataclass(data.config),
            enabled=data.enabled,
            created_at=data.created_at,
        )


@strawberry.type(description="Notification rule")
class NotificationRule(Node):
    id: NodeID[str]
    name: str
    description: Optional[str]
    rule_type: NotificationRuleTypeGQL
    channel: NotificationChannel
    message_template: str
    enabled: bool
    created_at: datetime

    @classmethod
    def from_dataclass(cls, data: NotificationRuleData) -> Self:
        return cls(
            id=ID(str(data.id)),
            name=data.name,
            description=data.description,
            rule_type=NotificationRuleTypeGQL.from_internal(data.rule_type),
            channel=NotificationChannel.from_dataclass(data.channel),
            message_template=data.message_template,
            enabled=data.enabled,
            created_at=data.created_at,
        )


# Filter and OrderBy types


@strawberry.enum
class NotificationChannelOrderField(StrEnum):
    NAME = "name"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


@strawberry.input(description="Filter for notification channels")
class NotificationChannelFilter(GQLFilter):
    name: Optional[StringFilter] = None
    channel_type: Optional[list[NotificationChannelTypeGQL]] = None
    enabled: Optional[bool] = None

    AND: Optional[list[NotificationChannelFilter]] = None
    OR: Optional[list[NotificationChannelFilter]] = None
    NOT: Optional[list[NotificationChannelFilter]] = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list containing a single combined QueryCondition that represents
        all filters with proper logical operators applied.
        """
        # Collect direct field conditions (these will be combined with AND)
        field_conditions: list[QueryCondition] = []

        # Apply name filter
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=NotificationChannelConditions.by_name_contains,
                equals_factory=NotificationChannelConditions.by_name_equals,
            )
            if name_condition:
                field_conditions.append(name_condition)

        # Apply channel_type filter
        if self.channel_type:
            internal_types = [ct.to_internal() for ct in self.channel_type]
            field_conditions.append(NotificationChannelConditions.by_channel_types(internal_types))

        # Apply enabled filter
        if self.enabled is not None:
            field_conditions.append(NotificationChannelConditions.by_enabled(self.enabled))

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(description="Order by specification for notification channels")
class NotificationChannelOrderBy(GQLOrderBy):
    field: NotificationChannelOrderField
    direction: OrderDirection = OrderDirection.ASC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case NotificationChannelOrderField.NAME:
                return NotificationChannelOrders.name(ascending)
            case NotificationChannelOrderField.CREATED_AT:
                return NotificationChannelOrders.created_at(ascending)
            case NotificationChannelOrderField.UPDATED_AT:
                return NotificationChannelOrders.updated_at(ascending)


@strawberry.enum
class NotificationRuleOrderField(StrEnum):
    NAME = "name"
    CREATED_AT = "created_at"
    UPDATED_AT = "updated_at"


@strawberry.input(description="Filter for notification rules")
class NotificationRuleFilter(GQLFilter):
    name: Optional[StringFilter] = None
    rule_type: Optional[list[NotificationRuleTypeGQL]] = None
    enabled: Optional[bool] = None

    AND: Optional[list[NotificationRuleFilter]] = None
    OR: Optional[list[NotificationRuleFilter]] = None
    NOT: Optional[list[NotificationRuleFilter]] = None

    @override
    def build_conditions(self) -> list[QueryCondition]:
        """Build query conditions from this filter.

        Returns a list containing a single combined QueryCondition that represents
        all filters with proper logical operators applied.
        """
        # Collect direct field conditions (these will be combined with AND)
        field_conditions: list[QueryCondition] = []

        # Apply name filter
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=NotificationRuleConditions.by_name_contains,
                equals_factory=NotificationRuleConditions.by_name_equals,
            )
            if name_condition:
                field_conditions.append(name_condition)

        # Apply rule_type filter
        if self.rule_type:
            internal_types = [rt.to_internal() for rt in self.rule_type]
            field_conditions.append(NotificationRuleConditions.by_rule_types(internal_types))

        # Apply enabled filter
        if self.enabled is not None:
            field_conditions.append(NotificationRuleConditions.by_enabled(self.enabled))

        # Handle AND logical operator - these are implicitly ANDed with field conditions
        if self.AND:
            for sub_filter in self.AND:
                field_conditions.extend(sub_filter.build_conditions())

        # Handle OR logical operator
        if self.OR:
            or_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_sub_conditions.extend(sub_filter.build_conditions())
            if or_sub_conditions:
                field_conditions.append(combine_conditions_or(or_sub_conditions))

        # Handle NOT logical operator
        if self.NOT:
            not_sub_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_sub_conditions.extend(sub_filter.build_conditions())
            if not_sub_conditions:
                field_conditions.append(negate_conditions(not_sub_conditions))

        return field_conditions


@strawberry.input(description="Order by specification for notification rules")
class NotificationRuleOrderBy(GQLOrderBy):
    field: NotificationRuleOrderField
    direction: OrderDirection = OrderDirection.ASC

    @override
    def to_query_order(self) -> QueryOrder:
        """Convert to repository QueryOrder."""
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case NotificationRuleOrderField.NAME:
                return NotificationRuleOrders.name(ascending)
            case NotificationRuleOrderField.CREATED_AT:
                return NotificationRuleOrders.created_at(ascending)
            case NotificationRuleOrderField.UPDATED_AT:
                return NotificationRuleOrders.updated_at(ascending)


# Input types for mutations


@strawberry.input(description="Input for webhook configuration")
class WebhookConfigInput:
    url: str

    def to_dataclass(self) -> WebhookConfig:
        return WebhookConfig(url=self.url)


@strawberry.input(description="Input for creating a notification channel")
class CreateNotificationChannelInput:
    name: str
    description: Optional[str] = None
    channel_type: NotificationChannelTypeGQL = NotificationChannelTypeGQL.WEBHOOK
    config: WebhookConfigInput = strawberry.field()
    enabled: bool = True

    def to_creator(self, created_by: uuid.UUID) -> NotificationChannelCreator:
        return NotificationChannelCreator(
            name=self.name,
            description=self.description,
            channel_type=self.channel_type.to_internal(),
            config=self.config.to_dataclass(),
            enabled=self.enabled,
            created_by=created_by,
        )


@strawberry.input(description="Input for updating a notification channel")
class UpdateNotificationChannelInput:
    id: ID
    name: Optional[str] = UNSET
    description: Optional[str] = UNSET
    config: Optional[WebhookConfigInput] = UNSET
    enabled: Optional[bool] = UNSET

    def to_modifier(self) -> NotificationChannelModifier:
        config_state = OptionalState[WebhookConfig].nop()
        if self.config is not UNSET:
            if self.config is None:
                config_state = OptionalState[WebhookConfig].nop()
            else:
                config_state = OptionalState[WebhookConfig].update(self.config.to_dataclass())

        return NotificationChannelModifier(
            name=OptionalState[str].from_graphql(self.name),
            description=OptionalState[Optional[str]].from_graphql(self.description),
            config=config_state,
            enabled=OptionalState[bool].from_graphql(self.enabled),
        )


@strawberry.input(description="Input for deleting a notification channel")
class DeleteNotificationChannelInput:
    id: ID


@strawberry.input(description="Input for creating a notification rule")
class CreateNotificationRuleInput:
    name: str
    description: Optional[str] = None
    rule_type: NotificationRuleTypeGQL = strawberry.field()
    channel_id: ID
    message_template: str
    enabled: bool = True

    def to_creator(self, created_by: uuid.UUID) -> NotificationRuleCreator:
        return NotificationRuleCreator(
            name=self.name,
            description=self.description,
            rule_type=self.rule_type.to_internal(),
            channel_id=uuid.UUID(self.channel_id),
            message_template=self.message_template,
            enabled=self.enabled,
            created_by=created_by,
        )


@strawberry.input(description="Input for updating a notification rule")
class UpdateNotificationRuleInput:
    id: ID
    name: Optional[str] = UNSET
    description: Optional[str] = UNSET
    message_template: Optional[str] = UNSET
    enabled: Optional[bool] = UNSET

    def to_modifier(self) -> NotificationRuleModifier:
        return NotificationRuleModifier(
            name=OptionalState[str].from_graphql(self.name),
            description=OptionalState[Optional[str]].from_graphql(self.description),
            message_template=OptionalState[str].from_graphql(self.message_template),
            enabled=OptionalState[bool].from_graphql(self.enabled),
        )


@strawberry.input(description="Input for deleting a notification rule")
class DeleteNotificationRuleInput:
    id: ID


# Payload types for mutations


@strawberry.type(description="Payload for create notification channel mutation")
class CreateNotificationChannelPayload:
    channel: NotificationChannel


@strawberry.type(description="Payload for update notification channel mutation")
class UpdateNotificationChannelPayload:
    channel: NotificationChannel


@strawberry.type(description="Payload for delete notification channel mutation")
class DeleteNotificationChannelPayload:
    id: ID


@strawberry.type(description="Payload for create notification rule mutation")
class CreateNotificationRulePayload:
    rule: NotificationRule


@strawberry.type(description="Payload for update notification rule mutation")
class UpdateNotificationRulePayload:
    rule: NotificationRule


@strawberry.type(description="Payload for delete notification rule mutation")
class DeleteNotificationRulePayload:
    id: ID


# Validate mutations


@strawberry.input(description="Input for validate notification channel mutation")
class ValidateNotificationChannelInput:
    id: ID
    test_message: str


@strawberry.type(description="Payload for validate notification channel mutation")
class ValidateNotificationChannelPayload:
    id: ID


@strawberry.input(description="Input for validate notification rule mutation")
class ValidateNotificationRuleInput:
    id: ID
    notification_data: Optional[strawberry.scalars.JSON] = UNSET


@strawberry.type(description="Payload for validate notification rule mutation")
class ValidateNotificationRulePayload:
    message: str



from __future__ import annotations

import logging
import uuid

import sqlalchemy as sa
from sqlalchemy.orm import foreign, relationship

from ai.backend.common.data.artifact.types import VerificationStepResult
from ai.backend.common.data.storage.registries.types import ModelData
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.artifact.types import (
    ArtifactRemoteStatus,
    ArtifactRevisionData,
    ArtifactStatus,
)
from ai.backend.manager.models.association_artifacts_storages import AssociationArtifactsStorageRow

from .base import (
    GUID,
    Base,
    IDColumn,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("ArtifactRevisionRow",)


def _get_artifact_join_cond():
    from .artifact import ArtifactRow

    return foreign(ArtifactRevisionRow.artifact_id) == ArtifactRow.id


class ArtifactRevisionRow(Base):
    __tablename__ = "artifact_revisions"
    __table_args__ = (
        # constraint
        sa.UniqueConstraint("artifact_id", "version", name="uq_artifact_id_version"),
    )

    id = IDColumn("id")
    artifact_id = sa.Column(
        GUID,
        nullable=False,
        index=True,
    )
    version = sa.Column("version", sa.String, nullable=False)
    readme = sa.Column("readme", sa.TEXT, nullable=True, default=None)
    size = sa.Column("size", sa.BigInteger, nullable=True, default=None)
    digest = sa.Column("digest", sa.String, nullable=True, server_default=None, default=None)

    # It's unnatural to include "status" in the revision, but let's put it here for now instead of creating separate table.
    status = sa.Column(sa.String, index=True, nullable=False, default=ArtifactStatus.SCANNED.value)
    remote_status = sa.Column(
        sa.String, index=True, nullable=True, default=None, server_default=sa.null()
    )

    created_at = sa.Column(
        "created_at",
        sa.DateTime(timezone=True),
        nullable=True,
        server_default=None,
        index=True,
    )
    updated_at = sa.Column(
        "updated_at",
        sa.DateTime(timezone=True),
        nullable=True,
        server_default=None,
        index=True,
    )
    verification_result = sa.Column(
        "verification_result", sa.JSON(none_as_null=True), nullable=True, default=None
    )

    artifact = relationship(
        "ArtifactRow",
        back_populates="revision_rows",
        primaryjoin=_get_artifact_join_cond,
        viewonly=True,
    )

    association_artifacts_storages_rows = relationship(
        "AssociationArtifactsStorageRow",
        back_populates="artifact_revision_row",
        primaryjoin=lambda: ArtifactRevisionRow.id
        == foreign(AssociationArtifactsStorageRow.artifact_revision_id),
    )

    def __str__(self) -> str:
        return (
            f"ArtifactRevisionRow("
            f"id={self.id}, "
            f"artifact_id={self.artifact_id}, "
            f"version={self.version}, "
            f"readme={self.readme[:15]}, "  # truncate for display
            f"size={self.size}, "
            f"status={self.status}, "
            f"remote_status={self.remote_status}, "
            f"created_at={self.created_at.isoformat()}, "
            f"updated_at={self.updated_at.isoformat()}, "
            f"digest={self.digest}"
            f")"
        )

    def to_dataclass(self) -> ArtifactRevisionData:
        # Convert JSON dict back to Pydantic model if present
        verification_result = None
        if self.verification_result is not None:
            try:
                verification_result = VerificationStepResult.model_validate(
                    self.verification_result
                )
            except Exception as e:
                # If validation fails, keep as None
                verification_result = None
                log.warning(
                    "Failed to validate verification_result for ArtifactRevisionRow id={}: {}",
                    self.id,
                    e,
                )

        return ArtifactRevisionData(
            id=self.id,
            artifact_id=self.artifact_id,
            version=self.version,
            readme=self.readme,
            size=self.size,
            status=ArtifactStatus(self.status),
            remote_status=ArtifactRemoteStatus(self.remote_status) if self.remote_status else None,
            created_at=self.created_at,
            updated_at=self.updated_at,
            digest=self.digest,
            verification_result=verification_result,
        )

    @classmethod
    def from_huggingface_model_data(
        cls,
        artifact_id: uuid.UUID,
        model_data: ModelData,
    ) -> ArtifactRevisionRow:
        return cls(
            artifact_id=artifact_id,
            version=model_data.revision,
            readme=model_data.readme,
            size=model_data.size,
            status=ArtifactStatus.SCANNED.value,
            remote_status=None,
            created_at=model_data.created_at,
            updated_at=model_data.modified_at,
            digest=model_data.sha,
            verification_result=None,
        )



"""Validators for the Sokovan scheduler."""

from .concurrency import ConcurrencyValidator
from .dependencies import DependenciesValidator
from .domain_resource_limit import DomainResourceLimitValidator
from .exceptions import (
    ConcurrencyLimitExceeded,
    DependenciesNotSatisfied,
    DomainResourceQuotaExceeded,
    GroupResourceQuotaExceeded,
    KeypairResourceQuotaExceeded,
    PendingSessionCountLimitExceeded,
    PendingSessionResourceLimitExceeded,
    SchedulingValidationError,
    UserResourceQuotaExceeded,
)
from .group_resource_limit import GroupResourceLimitValidator
from .keypair_resource_limit import KeypairResourceLimitValidator
from .pending_session_count_limit import PendingSessionCountLimitValidator
from .pending_session_resource_limit import PendingSessionResourceLimitValidator
from .reserved_batch import ReservedBatchSessionValidator
from .user_resource_limit import UserResourceLimitValidator
from .validator import SchedulingValidator, ValidatorRule

__all__ = [
    "ConcurrencyValidator",
    "DependenciesValidator",
    "DomainResourceLimitValidator",
    "GroupResourceLimitValidator",
    "KeypairResourceLimitValidator",
    "PendingSessionCountLimitValidator",
    "PendingSessionResourceLimitValidator",
    "ReservedBatchSessionValidator",
    "UserResourceLimitValidator",
    "SchedulingValidator",
    "ValidatorRule",
    "SchedulingValidationError",
    "ConcurrencyLimitExceeded",
    "DependenciesNotSatisfied",
    "KeypairResourceQuotaExceeded",
    "UserResourceQuotaExceeded",
    "GroupResourceQuotaExceeded",
    "DomainResourceQuotaExceeded",
    "PendingSessionCountLimitExceeded",
    "PendingSessionResourceLimitExceeded",
]



import uuid
from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.services.group.actions.base import GroupAction


@dataclass
class DeleteGroupAction(GroupAction):
    group_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete"


@dataclass
class DeleteGroupActionResult(BaseActionResult):
    group_id: uuid.UUID

    @override
    def entity_id(self) -> Optional[str]:
        return str(self.group_id)



from dataclasses import dataclass
from typing import Optional, override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.data.domain.types import UserInfo
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class DeleteDomainAction(DomainAction):
    name: str
    user_info: UserInfo

    @override
    def entity_id(self) -> Optional[str]:
        return None

    @override
    @classmethod
    def operation_type(cls) -> str:
        return "delete"


@dataclass
class DeleteDomainActionResult(BaseActionResult):
    success: bool
    description: str

    @override
    def entity_id(self) -> Optional[str]:
        return None
